'''
Read pixels from one or more consecutive regions
and group them into hexagonal bins
Assume the reader reads a file containing columns gene, X, Y
If input contains a column provided by region_id, each region is processed separately, regions do not have to be consecutive or be in the same coordinate system
'''
import sys, os
import numpy as np
import pandas as pd
from scipy.sparse import coo_array, vstack

from ficture.utils.hexagon_fn import *

class PixelToUnit:

    def __init__(self, reader, ft_dict, key, radius, scale=1, region_id=None, min_ct_per_unit=1, sliding_step=1, major_axis=None, xy_lattice=False) -> None:
        self.reader = reader
        self.ft_dict = ft_dict
        self.M = len(self.ft_dict)
        self.key = key
        self.radius = radius
        self.file_is_open = True
        self.min_ct_per_unit = min_ct_per_unit
        self.scale = scale
        self.region_id = region_id
        self.n_move = sliding_step
        self.Y = major_axis
        self.mj = self.Y if self.Y is not None else 'X'
        self.mi = 'X' if self.mj == 'Y' else 'Y'
        self.df = pd.DataFrame()
        self.brc = None
        self.mtx = None
        self.lattice = xy_lattice
        self.n_batch = 0
        self.bound = self.radius * np.sqrt(3)/2

    def read_chunk(self, min_size = 200):
        if not self.file_is_open:
            return 0
        if self.region_id is None:
            return self._read_chunk_consecutive(min_size)
        else:
            return self._read_chunk_region(min_size)

    def _read_chunk_region(self, min_size = 200):
        # todo: test this function
        if not self.file_is_open:
            return -1
        mj_range = [-np.inf, np.inf]
        mi_range = [-np.inf, np.inf]
        mj_size = 0
        mi_size = 0
        region_list = []
        prev_region = None
        if len(self.df) > 0:
            mj_range = [self.df[self.mj].max(), self.df[self.mj].min()]
            mi_range = [self.df[self.mi].max(), self.df[self.mi].min()]
            mj_size = mj_range[0] - mj_range[1]
            mi_size = mi_range[0] - mi_range[1]
            region_list = list(self.df[self.region_id].unique())
            prev_region = self.df[self.region_id].iloc[-1]
        while (mj_size < min_size or mi_size < min_size) and len(region_list) < 2:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            chunk = chunk[chunk.gene.isin(self.ft_dict)]
            chunk.X = chunk.X.astype(float) * self.scale
            chunk.Y = chunk.Y.astype(float) * self.scale
            last_region = chunk[self.region_id].iloc[-1]
            indx = chunk[self.region_id].eq(last_region)
            if last_region == prev_region:
                mj_range[0] = max(mj_range[0], chunk.loc[indx, self.mj].max())
                mj_range[1] = min(mj_range[1], chunk.loc[indx, self.mj].min())
                mi_range[0] = max(mi_range[0], chunk.loc[indx, self.mi].max())
                mi_range[1] = min(mi_range[1], chunk.loc[indx, self.mi].min())
            else:
                mj_range = [chunk.loc[indx, self.mj].max(), chunk.loc[indx, self.mj].min()]
                mi_range = [chunk.loc[indx, self.mi].max(), chunk.loc[indx, self.mi].min()]
            mj_size = mj_range[0] - mj_range[1]
            mi_size = mi_range[0] - mi_range[1]
            prev_region = last_region
            region_list += [x for x in chunk[self.region_id].unique() if x not in region_list]
            self.df = pd.concat([self.df, chunk])
        left = pd.DataFrame()
        if len(region_list) > 1 and (mj_size < min_size or mi_size < min_size):
            left = self.df[self.df[self.region_id].eq(region_list[-1])]
            region_list = region_list[:-1]
        elif self.Y is not None:
            left = self.df.loc[self.df[self.region_id].eq(region_list[-1]) & (self.df[self.mj] > mj_range[1])]
        self.df['hex_id'] = ''
        self.brc = pd.DataFrame()
        self.mtx = coo_array(([], ([], [])), shape = (0, self.M))
        for reg in region_list:
            indx = self.df[self.region_id].eq(reg)
            for offs_x in range(self.n_move):
                for offs_y in range(self.n_move):
                    offs_iden = str(offs_x) + '_' + str(offs_y)
                    x, y = pixel_to_hex(self.df.loc[indx, ['X', 'Y']].values, self.radius, offs_x/self.n_move, offs_y/self.n_move)
                    self.df.loc[indx, "hex_id"] = list(zip(x, y))
                    ct = self.df[indx].groupby('hex_id').agg({self.key: "sum"}).reset_index()
                    ct.drop(index = ct.index[ct[self.key] < self.min_ct_per_unit], inplace=True)
                    if len(ct) < 1:
                        continue
                    bt_dict = {x:i for i,x in enumerate(ct.hex_id.values)}
                    N = len(bt_dict)
                    brc = self.df[indx & self.df.hex_id.isin(bt_dict)].groupby(by = ['hex_id', 'gene']).agg({self.key: "sum"}).reset_index()
                    self.mtx = vstack([self.mtx, coo_array((brc[self.key].values, (brc.hex_id.map(bt_dict).values, brc.gene.map(self.ft_dict).values) ), shape = (N, self.M))])
                    if self.lattice:
                        ct['x'], ct['y'] = \
                            hex_to_pixel([v[0] for v in ct.hex_id.values],\
                                         [v[1] for v in ct.hex_id.values],\
                            self.radius, offs_x/self.n_move, offs_y/self.n_move)
                    else:
                        ct = brc.groupby(by = ['hex_id']).agg({self.key: sum,\
                            'X': np.median, 'Y':np.median}).reset_index()
                        ct.rename(columns = {'X':'x', 'Y':'y'}, inplace=True)
                    ct.index = ct.hex_id.map(bt_dict)
                    ct.sort_index(inplace=True)
                    ct['hex_id'] = ct.hex_id.map(lambda x : '_'.join([str(u) for u in x]))
                    ct.hex_id = offs_iden + '_' + ct.hex_id.values
                    ct[self.region_id] = reg
                    self.brc = pd.concat([self.brc, ct])
        self.mtx = self.mtx.tocsr()
        self.df = left
        self.brc.index = range(self.brc.shape[0])
        return self.brc.shape[0]


    def _read_chunk_consecutive(self, min_size = 200):
        if not self.file_is_open:
            return -1
        mj_range = [-np.inf, np.inf]
        mi_range = [-np.inf, np.inf]
        mj_size = 0
        mi_size = 0
        if len(self.df) > 0:
            mj_range = [self.df[self.mj].iloc[-1], self.df[self.mj].iloc[0]]
            mi_range = [self.df[self.mi].max(), self.df[self.mi].min()]
            mj_size = mj_range[0] - mj_range[1]
            mi_size = mi_range[0] - mi_range[1]
        while mj_size < min_size or mi_size < min_size:
            try:
                chunk = next(self.reader)
            except StopIteration:
                self.file_is_open = False
                break
            chunk = chunk[chunk.gene.isin(self.ft_dict)]
            chunk.X = chunk.X.astype(float) * self.scale
            chunk.Y = chunk.Y.astype(float) * self.scale
            mj_range[0] = chunk[self.mj].iloc[-1]
            mj_range[1] = min(mj_range[1], chunk[self.mj].iloc[0])
            mi_range[0] = max(mi_range[0], chunk[self.mi].max())
            mi_range[1] = min(mi_range[1], chunk[self.mi].min())
            mj_size = mj_range[0] - mj_range[1]
            mi_size = mi_range[0] - mi_range[1]
            self.df = pd.concat([self.df, chunk])
        self.df['hex_id'] = ''
        self.brc = pd.DataFrame()
        self.mtx = coo_array(([], ([], [])), shape = (0, self.M))
        for offs_x in range(self.n_move):
            for offs_y in range(self.n_move):
                offs_iden = str(offs_x) + '_' + str(offs_y)
                x, y = pixel_to_hex(self.df[['X', 'Y']].values, self.radius, offs_x/self.n_move, offs_y/self.n_move)
                self.df.hex_id = list(zip(x, y))
                ct = self.df.groupby('hex_id').agg({self.key: "sum"}).reset_index()
                ct['X'], ct['Y'] = \
                        hex_to_pixel([v[0] for v in ct.hex_id.values],\
                                     [v[1] for v in ct.hex_id.values],\
                        self.radius, offs_x/self.n_move, offs_y/self.n_move)
                # Drop hexagons with centers within sqrt(3)*raidus/2 from each boundary
                ct.drop(index = ct.index[\
                    (ct[self.mj].lt(mj_range[1] + self.bound) ) | \
                    (ct[self.mj].gt(mj_range[0] - self.bound) ) ], inplace=True)
                ct.drop(index = ct.index[ct[self.key] < self.min_ct_per_unit], inplace=True)
                if len(ct) < 1:
                    continue
                bt_dict = {x:i for i,x in enumerate(ct.hex_id.values)}
                N = len(bt_dict)
                brc = self.df[self.df.hex_id.isin(bt_dict)].groupby(by = ['hex_id', 'gene']).agg({self.key: "sum"}).reset_index()
                self.mtx = vstack([self.mtx, coo_array((brc[self.key].values, (brc.hex_id.map(bt_dict).values, brc.gene.map(self.ft_dict).values) ), shape = (N, self.M))])
                if not self.lattice:
                    ct = brc.groupby(by = ['hex_id']).agg({self.key: sum, 'X': np.median, 'Y':np.median}).reset_index()
                ct.rename(columns = {'X':'x', 'Y':'y'}, inplace=True)
                ct.index = ct.hex_id.map(bt_dict)
                ct.sort_index(inplace=True)
                ct['hex_id'] = ct.hex_id.map(lambda x : '_'.join([str(u) for u in x]))
                ct.hex_id = offs_iden + '_' + ct.hex_id.values
                self.brc = pd.concat([self.brc, ct])
        self.df = self.df[self.df[self.mj] >= mj_range[0] - 2 * self.radius]
        self.df.drop(columns = ['hex_id'], inplace = True)
        self.mtx = self.mtx.tocsr()
        self.brc.index = range(self.brc.shape[0])
        self.n_batch += 1
        return self.brc.shape[0]



import sys, os, gzip, argparse, logging, warnings, shutil

from ficture.utils.minimake import minimake

def run_together(_args):
    """Run all functions in FICTURE by using GNU Makefile
    This function is meant to be used in a local environment that has sufficient resources to run all functions in FICTURE at once.
    This function performs the following tasks:
    (1) Take the input parameters relevant to the FICTURE runs
    (2) Identify the sequence of commands to run FICTURE
    (3) Create a GNU makefile to run the commands in parallel
    (4) Run the GNU makefile
    """

    parser = argparse.ArgumentParser()

    cmd_params = parser.add_argument_group("Commands", "FICTURE commands to run together")
    cmd_params.add_argument('--all', action='store_true', default=False, help='Run all FICTURE commands (preprocess, segment, lda, decode)')
    cmd_params.add_argument('--preprocess', action='store_true', default=False, help='Perform preprocess step')
    cmd_params.add_argument('--segment', action='store_true', default=False, help='Perform hexagonal segmentation')
    cmd_params.add_argument('--sge', action='store_true', default=False, help='Create hexagonal SGEs')
    cmd_params.add_argument('--lda', action='store_true', default=False, help='Perform LDA model training')
    cmd_params.add_argument('--decode', action='store_true', default=False, help='Perform pixel-level decoding')

    run_params = parser.add_argument_group("Run Options", "Run options for FICTURE commands")
    run_params.add_argument('--dry-run', action='store_true', default=False, help='Dry run. Generate only the Makefile without running it')
    run_params.add_argument('--restart', action='store_true', default=False, help='Restart the run. Ignore all intermediate files and start from the beginning')
    run_params.add_argument('--threads', type=int, default=1, help='Maximum number of threads to use in each process')
    run_params.add_argument('--n-jobs', type=int, default=1, help='Number of jobs (processes) to run in parallel')

    key_params = parser.add_argument_group("Key Parameters", "Key parameters that requires user's attention")
    key_params.add_argument('--in-tsv', required=True, type=str, help='Input TSV file (e.g. transcript.tsv.gz)')
    key_params.add_argument('--out-dir', required= True, type=str, help='Output directory')
    key_params.add_argument('--in-minmax', type=str, help='Input coordinate minmax TSV file (e.g. coordinate_minmax.tsv). If absent, it will be generated')
    key_params.add_argument('--in-feature', type=str, help='Input TSV file (e.g. feature.clean.tsv.gz) that specify which genes to use as input. If absent, it will be use all genes')
    key_params.add_argument('--major-axis', type=str, default='Y', help='Axis where transcripts.tsv.gz are sorted')
    key_params.add_argument('--mu-scale', type=float, default=1.0, help='Scale factor for mu (pixels per um)')
    key_params.add_argument('--train-width', type=str, default="12", help='Hexagon flat-to-flat width (in um) during training. Use comma to specify multiple values')
    key_params.add_argument('--n-factor', type=str, default="12", help='Number of factors to train. Use comma to specify multiple values')
    key_params.add_argument('--anchor-res', type=float, default=4, help='Anchor resolution for decoding')
    key_params.add_argument('--plot-each-factor', action='store_true', default=False, help='Plot individual factors in pixel-level decoding')

    aux_params = parser.add_argument_group("Auxiliary Parameters", "Auxiliary parameters (using default is recommended)")
    aux_params.add_argument('--train-epoch', type=int, default=3, help='Training epoch for LDA model')
    aux_params.add_argument('--train-epoch-id-len', type=int, default=2, help='Training epoch ID length')
    aux_params.add_argument('--train-n-move', type=int, default=2, help='Level of hexagonal sliding during training')
    aux_params.add_argument('--sge-n-move', type=int, default=1, help='Level of hexagonal sliding during SGE generation')
    aux_params.add_argument('--fit-width', type=float, help='Hexagon flat-to-flat width (in um) during model fitting (default: same to train-width)')
    aux_params.add_argument('--key-col', type=str, default="Count", help='Columns from the input file to be used as key')
    aux_params.add_argument('--minibatch-size', type=int, default=500, help='Batch size used in minibatch processing')
    aux_params.add_argument('--minibatch-buffer', type=int, default=30, help='Batch buffer used in minibatch processing')
    aux_params.add_argument('--min-ct-unit-dge', type=int, default=50, help='Minimum count per hexagon in DGE generation')
    aux_params.add_argument('--min-ct-unit-sge', type=int, default=1, help='Minimum count per hexagon in SGE generation')
    aux_params.add_argument('--min-ct-feature', type=int, default=20, help='Minimum count per feature during LDA training')
    aux_params.add_argument('--min-ct-unit-fit', type=int, default=20, help='Minimum count per hexagon unit during model fitting')
    aux_params.add_argument('--lda-rand-init', type=int, default=10, help='Number of random initialization during model training')
    aux_params.add_argument('--decode-top-k', type=int, default=3, help='Top K columns to output in pixel-level decoding results')
    aux_params.add_argument('--de-max-pval', type=float, default=1e-3, help='p-value cutoff for differential expression')
    aux_params.add_argument('--de-min-fold', type=float, default=1.5, help='Fold-change cutoff for differential expression')
    aux_params.add_argument('--decode-block-size', type=int, default=100, help='Block size for pixel decoding output')
    aux_params.add_argument('--decode-scale', type=int, default=100, help='Scale parameters for pixel decoding output')
    aux_params.add_argument('--cmap-name', type=str, default="turbo", help='Name of color map')
    aux_params.add_argument('--dge-precision', type=float, default=2, help='Output precision of hexagon coordinates')
    aux_params.add_argument('--fit-precision', type=float, default=2, help='Output precision of model fitting')
    aux_params.add_argument('--decode-precision', type=float, default=0.1, help='Precision of pixel level decoding')
    aux_params.add_argument('--lda-plot-um-per-pixel', type=float, default=1, help='Image resolution for LDA plot')
    aux_params.add_argument('--decode-plot-um-per-pixel', type=float, default=0.5, help='Image resolution for pixel decoding plot')
    aux_params.add_argument('--decode-sub-um-per-pixel', type=float, default=1, help='Image resolution for individual subplots')
    # aux_params.add_argument('--bgzip', type=str, default="bgzip", help='Path to bgzip binary. For faster processing, use "bgzip -@ 4')
    # aux_params.add_argument('--tabix', type=str, default="tabix", help='Path to tabix binary')
    aux_params.add_argument('--gzip', type=str, default="gzip", help='Path to gzip binary. For faster processing, use "pigz -p 4"')
    aux_params.add_argument('--sort', type=str, default="sort", help='Path to sort binary. For faster processing, you may add arguments like "sort -T /path/to/new/tmpdir --parallel=20 -S 10G"')

    args = parser.parse_args(_args)

    if len(_args) == 0:
        parser.print_help()
        return

    if args.all:
        args.preprocess = True
        args.segment = True
        args.sge = False
        args.lda = True
        args.decode = True

    ## parse input parameters
    train_widths = [int(x) for x in args.train_width.split(",")]
    n_factors = [int(x) for x in args.n_factor.split(",")]

    batch_tsv = f"{args.out_dir}/batched.matrix.tsv"
    batch_out = f"{args.out_dir}/batched.matrix.tsv.gz"
    minmax_out = args.in_minmax if args.in_minmax is not None else f"{args.out_dir}/coordinate_minmax.tsv"

    ## create output directory
    os.makedirs(args.out_dir, exist_ok=True)

    mm = minimake()

    ## gzip and sort have to exist across everywhere..
    if not shutil.which(args.gzip.split(" ")[0]):
        logging.error(f"Cannot find {args.gzip}. Please make sure that the path to --gzip is correct")
        sys.exit(1)

    if not shutil.which(args.sort.split(" ")[0]):
        logging.error(f"Cannot find {args.sort}. Please make sure that the path to --sort is correct")
        sys.exit(1)

    if args.preprocess:
        ## create output directory if needed
        cmds = []
        cmds.append(rf"$(info --------------------------------------------------------------)")
        cmds.append(rf"$(info Creating minibatch from {args.in_tsv}...)")
        cmds.append(rf"$(info --------------------------------------------------------------)")
        ## create minibatch
        cmds.append(f"ficture make_spatial_minibatch --input {args.in_tsv} --output {batch_tsv} --mu_scale {args.mu_scale} --batch_size {args.minibatch_size} --batch_buff {args.minibatch_buffer} --major_axis {args.major_axis}")
        # cmds.append(f"{args.sort} -k 2,2n -k 1,1g {batch_tsv} | {args.gzip} -c > {batch_out}")
        cmds.append(f"{args.gzip} -f {batch_tsv}")
        # cmds.append(f"rm {batch_tsv}")

        mm.add_target(batch_out, [args.in_tsv], cmds)

        if args.in_minmax is None:
            script_path = f"{args.out_dir}/write_minmax.sh"
            with open(script_path, "w") as f:
                f.write(r"""#!/bin/bash
input=$1
output=$2
mu_scale=$3
gzip -cd ${input} | awk 'BEGIN{FS=OFS="\t"} NR==1{for(i=1;i<=NF;i++){if($i=="X")x=i;if($i=="Y")y=i}print $x,$y;next}{print $x,$y}' | perl -slane 'print join("\t",$F[0]/${mu_scale},$F[1]/${mu_scale})' -- -mu_scale="${mu_scale}" | awk -F'\t' ' BEGIN { min1 = "undef"; max1 = "undef"; min2 = "undef"; max2 = "undef"; } { if (NR == 2 || $1 < min1) min1 = $1; if (NR == 2 || $1 > max1) max1 = $1; if (NR == 2 || $2 < min2) min2 = $2; if (NR == 2 || $2 > max2) max2 = $2; } END { print "xmin\t", min1; print "xmax\t", max1; print "ymin\t", min2; print "ymax\t", max2; }' > ${output}
""")
            cmds = []
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(rf"$(info Obtaining boundary coordinates to {minmax_out}...)")
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(f"bash {script_path} {args.in_tsv} {minmax_out} {args.mu_scale}")
            mm.add_target(minmax_out, [args.in_tsv], cmds)

    if args.segment:
        for train_width in train_widths:
            dge_out = f"{args.out_dir}/hexagon.d_{train_width}.tsv"
            cmds = []
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(rf"$(info Creating DGE for {train_width}um...)")
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(f"ficture make_dge --key {args.key_col} --input {args.in_tsv} --output {dge_out} --hex_width {train_width} --n_move {args.train_n_move} --min_ct_per_unit {args.min_ct_unit_dge} --mu_scale {args.mu_scale} --precision {args.dge_precision} --major_axis {args.major_axis}")
            cmds.append(f"{args.sort} -k 1,1n {dge_out} | {args.gzip} -c > {dge_out}.gz")
            cmds.append(f"rm {dge_out}")
            mm.add_target(f"{dge_out}.gz", [args.in_tsv], cmds)

    if args.sge:
        for train_width in train_widths:
            sge_out_dir = f"{args.out_dir}/segment/sge.d_{train_width}"
            cmds = []
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(rf"$(info Creating SGE for {train_width}um...)")
            cmds.append(rf"$(info --------------------------------------------------------------)")
            cmds.append(f"ficture make_sge_by_hexagon --key {args.key_col} --input {args.in_tsv} --feature {args.in_feature} --output_path {sge_out_dir} --hex_width {train_width} --n_move {args.sge_n_move} --min_ct_per_unit {args.min_ct_unit_sge} --mu_scale {args.mu_scale} --precision {args.dge_precision} --major_axis {args.major_axis} --transfer_gene_prefix")
            mm.add_target(f"{sge_out_dir}/barcodes.tsv.gz", [args.in_tsv], cmds)

    if args.lda:
        for train_width in train_widths:
            for n_factor in n_factors:
                model_id=f"nF{n_factor}.d_{train_width}"
                model_path=f"{args.out_dir}/analysis/{model_id}"
                figure_path=f"{model_path}/figure"
                hexagon = f"{args.out_dir}/hexagon.d_{train_width}.tsv.gz"
                model_prefix=f"{model_path}/{model_id}"
                model=f"{model_prefix}.model.p"
                feature_arg = f"--feature {args.in_feature}" if args.in_feature is not None else ""

                cmds = []
                cmds.append(rf"$(info --------------------------------------------------------------)")
                cmds.append(rf"$(info Creating LDA for {train_width}um and {n_factor} factors...)")
                cmds.append(rf"$(info --------------------------------------------------------------)")
                cmds.append(f"mkdir -p {model_path}/figure")
                cmds.append(f"ficture fit_model --input {hexagon} --output {model_prefix} {feature_arg} --nFactor {n_factor} --epoch {args.train_epoch} --epoch_id_length {args.train_epoch_id_len} --unit_attr X Y --key {args.key_col} --min_ct_per_feature {args.min_ct_feature} --test_split 0.5 --R {args.lda_rand_init} --thread {args.threads}")

                fit_tsv=f"{model_path}/{model_id}.fit_result.tsv.gz"
                fig_prefix=f"{figure_path}/{model_id}"
                cmap=f"{figure_path}/{model_id}.rgb.tsv"
                cmds.append(f"ficture choose_color --input {fit_tsv} --output {fig_prefix} --cmap_name {args.cmap_name}")

                fillr = (train_width / 2 + 1)
                cmds.append(f"ficture plot_base --input {fit_tsv} --output {fig_prefix}.coarse --fill_range {fillr} --color_table {cmap} --plot_um_per_pixel {args.lda_plot_um_per_pixel} --plot_discretized")
                cmds.append(f"touch {model_prefix}.done")

                mm.add_target(f"{model_prefix}.done", [args.in_tsv, hexagon], cmds)

    if args.decode:
        # if not shutil.which(args.bgzip.split(" ")[0]):
        #     logging.error(f"Cannot find {args.bgzip}. Please make sure that the path to --bgzip is correct")
        #     sys.exit(1)

        # if not shutil.which(args.tabix.split(" ")[0]):
        #     logging.error(f"Cannot find {args.tabix}. Please make sure that the path to --tabix is correct")
        #     sys.exit(1)


        script_path = f"{args.out_dir}/sort_decode.sh"
        with open(script_path, "w") as f:
            f.write(r"""#!/bin/bash
input=$1
output=$2
coor=$3
model_id=$4
bsize=$5
scale=$6
topk=$7
gzip=$8
#bgzip=$8
#tabix=$9

K=$( echo $model_id | sed 's/nF\([0-9]\{1,\}\)\..*/\1/' )
while IFS=$'\t' read -r r_key r_val; do
    export "${r_key}"="${r_val}"
done < ${coor}
echo -e "${xmin}, ${xmax}; ${ymin}, ${ymax}"

offsetx=${xmin}
offsety=${ymin}
rangex=$( echo "(${xmax} - ${xmin} + 0.5)/1+1" | bc )
rangey=$( echo "(${ymax} - ${ymin} + 0.5)/1+1" | bc )
bsize=2000
scale=100
header="##K=${K};TOPK=${topk}\n##BLOCK_SIZE=${bsize};BLOCK_AXIS=X;INDEX_AXIS=Y\n##OFFSET_X=${offsetx};OFFSET_Y=${offsety};SIZE_X=${rangex};SIZE_Y=${rangey};SCALE=${scale}\n#BLOCK\tX\tY\tK1\tK2\tK3\tP1\tP2\tP3"

(echo -e "${header}" && gzip -cd "${input}" | tail -n +2 | perl -slane '$F[0]=int(($F[1]-$offx)/$bsize) * $bsize; $F[1]=int(($F[1]-$offx)*$scale); $F[1]=($F[1]>=0)?$F[1]:0; $F[2]=int(($F[2]-$offy)*$scale); $F[2]=($F[2]>=0)?$F[2]:0; print join("\t", @F);' -- -bsize="${bsize}" -scale="${scale}" -offx="${offsetx}" -offy="${offsety}" | sort -S 1G -k1,1g -k3,3g ) | ${gzip} -c > ${output}

#${tabix} -f -s1 -b3 -e3 ${output}
rm ${input}
""")
        for train_width in train_widths:
            for n_factor in n_factors:
                batch_in = f"{args.out_dir}/batched.matrix.tsv.gz"
                model_id=f"nF{n_factor}.d_{train_width}"
                model_path=f"{args.out_dir}/analysis/{model_id}"
                figure_path=f"{model_path}/figure"
                model_prefix=f"{model_path}/{model_id}"
                cmap=f"{figure_path}/{model_id}.rgb.tsv"
                model=f"{args.out_dir}/analysis/{model_id}/{model_id}.model.p"

                if args.fit_width is None:
                    fit_widths = [train_width]
                else:
                    fit_widths = [float(x) for x in args.fit_width.split(",")]
                for fit_width in fit_widths:
                    cmds = []

                    fit_nmove = int(fit_width / args.anchor_res)
                    anchor_info=f"prj_{fit_width}.r_{args.anchor_res}"
                    radius = args.anchor_res + 1

                    prj_prefix = f"{model_path}/{model_id}.{anchor_info}"
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(rf"$(info Creating projection for {train_width}um and {n_factor} factors, at {fit_width}um)")
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(f"ficture transform --input {args.in_tsv} --output_pref {prj_prefix} --model {model} --key {args.key_col} --major_axis {args.major_axis} --hex_width {fit_width} --n_move {fit_nmove} --min_ct_per_unit {args.min_ct_unit_fit} --mu_scale {args.mu_scale} --thread {args.threads} --precision {args.fit_precision}")

                    batch_input=f"{args.out_dir}/batched.matrix.tsv.gz"
                    anchor=f"{prj_prefix}.fit_result.tsv.gz"
                    decode_basename=f"{model_id}.decode.{anchor_info}_{radius}"
                    decode_prefix=f"{model_path}/{decode_basename}"

                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(rf"$(info Performing pixel-level decoding..)")
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(f"ficture slda_decode --input {batch_in} --output {decode_prefix} --model {model} --anchor {anchor} --anchor_in_um --neighbor_radius {radius} --mu_scale {args.mu_scale} --key {args.key_col} --precision {args.decode_precision} --lite_topk_output_pixel {args.decode_top_k} --lite_topk_output_anchor {args.decode_top_k} --thread {args.threads}")

                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(rf"$(info Sorting and reformatting the pixel-level output..)")
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(f"bash {script_path} {decode_prefix}.pixel.tsv.gz {decode_prefix}.pixel.sorted.tsv.gz {minmax_out} {model_id} {args.decode_block_size} {args.decode_scale} {args.decode_top_k} {arg.gzip}")

                    de_input=f"{decode_prefix}.posterior.count.tsv.gz"
                    de_output=f"{decode_prefix}.bulk_chisq.tsv"

                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(rf"$(info Performing pseudo-bulk differential expression analysis..)")
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(f"ficture de_bulk --input {de_input} --output {de_output} --min_ct_per_feature {args.min_ct_feature} --max_pval_output {args.de_max_pval} --min_fold_output {args.de_min_fold}")

                    cmap=f"{figure_path}/{model_id}.rgb.tsv"
                    cmds.append(f"ficture factor_report --path {model_path} --pref {decode_basename} --color_table {cmap}")

                    decode_tsv=f"{decode_prefix}.pixel.sorted.tsv.gz"
                    decode_png=f"{model_path}/figure/{decode_basename}.pixel.png"

                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(rf"$(info Drawing pixel-level output image...)")
                    cmds.append(rf"$(info --------------------------------------------------------------)")
                    cmds.append(f"ficture plot_pixel_full --input {decode_tsv} --color_table {cmap} --output {decode_png} --plot_um_per_pixel {args.decode_plot_um_per_pixel} --full")

                    if args.plot_each_factor:
                        sub_prefix=f"{model_path}/figure/sub/{decode_basename}.pixel"
                        cmds.append(f"mkdir -p {model_path}/figure/sub")
                        cmds.append(f"ficture plot_pixel_single --input {decode_tsv} --output {sub_prefix} --plot_um_per_pixel {args.decode_sub_um_per_pixel} --full --all")

                    cmds.append(f"touch {decode_prefix}.done")
                    mm.add_target(f"{decode_prefix}.done", [batch_in, hexagon,f"{model_prefix}.done"], cmds)


    if len(mm.targets) == 0:
        logging.error("There is no target to run. Please make sure that at least on run option was turned on")
        sys.exit(1)

    ## write makefile
    mm.write_makefile(f"{args.out_dir}/Makefile")

    if args.dry_run:
        ## run makefile
        os.system(f"make -f {args.out_dir}/Makefile -n")
        print(f"To execute the pipeline, run the following command:\nmake -f {args.out_dir}/Makefile -j {args.n_jobs}")
    else:
        os.system(f"make -f {args.out_dir}/Makefile -j {args.n_jobs}")

if __name__ == "__main__":
    run_together(sys.argv[1:])



import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
import matplotlib.colors
from jinja2 import Environment, FileSystemLoader
from importlib import resources as impresources

import ficture.scripts
from ficture.utils.visualize_factors import visual_hc, image_to_base64

def factor_report(_args):

    parser = argparse.ArgumentParser(prog="factor_report")
    parser.add_argument('--path', type=str, help='')
    parser.add_argument('--pref', type=str, help='')
    parser.add_argument('--model_id', type=str, default='', help='')
    parser.add_argument('--color_table', type=str, default='', help='')
    parser.add_argument('--n_top_gene', type=int, default=20, help='')
    parser.add_argument('--min_top_gene', type=int, default=10, help='')
    parser.add_argument('--max_pval', type=float, default=0.001, help='')
    parser.add_argument('--min_fc', type=float, default=1.5, help='')
    parser.add_argument('--output_pref', type=str, default='', help='')
    parser.add_argument('--annotation', type=str, default = '', help='')

    parser.add_argument('--hc_tree', action='store_true')
    parser.add_argument('--n_top_gene_on_tree', type=int, default=10, help='')
    parser.add_argument('--tree_figure', type=str, default='', help='')
    parser.add_argument('--cprob_cut', type=str, default='.99', help='Only visualize top factors with cumulative probability > cprob_cut')
    parser.add_argument('--model', type=str, default='', help='')
    parser.add_argument('--anchor', type=str, default='', help='')
    parser.add_argument('--circle_if', type=int, default=24, help='')
    parser.add_argument('--remake_tree', action='store_true')
    parser.add_argument('--circle', action='store_true')
    parser.add_argument('--vertical', action='store_true')
    args = parser.parse_args(_args)

    if len(_args) == 0:
        parser.print_help()
        return

    path=args.path
    pref=args.pref
    ntop = args.n_top_gene
    mtop = args.min_top_gene
    pval_max = args.max_pval
    fc_min = args.min_fc
    ejs = os.path.join(impresources.files(ficture.scripts), "factor_report.template.html")
    if not os.path.isfile(ejs):
        sys.exit(f"Template file {ejs} not found")

    if args.remake_tree or args.circle or args.vertical:
        args.hc_tree = True

    model_id = args.model_id
    if model_id == '':
        model_id = args.pref

    output_pref = args.output_pref
    if output_pref == '':
        output_pref = path+"/"+pref

    # Color code
    color_f = args.color_table
    if not os.path.isfile(color_f):
        color_f = path+"/figure/"+model_id+".rgb.tsv"
    if not os.path.isfile(color_f):
        sys.exit(f"Cannot find color table")
    color_table = pd.read_csv(color_f, sep='\t')
    K = color_table.shape[0]
    logging.info(f"Read color table from {color_f}")

    factor_header = np.arange(K).astype(str)
    factor_name = {}
    if os.path.isfile(args.annotation):
        with open(args.annotation) as f:
            for line in f:
                x = line.strip().split('\t')
                factor_name[x[0]] = x[1]
                factor_header[int(x[0])] = x[1]

    print(factor_header)
    color_table['RGB'] = [','.join(x) for x in np.clip((color_table.loc[:, ['R','G','B']].values * 255).astype(int), 0, 255).astype(str) ]
    color_table['HEX'] = [ matplotlib.colors.to_hex(v) for v in np.array(color_table.loc[:, ['R','G','B']]) ]
    node_color = {factor_header[v['Name']]:v['HEX'] for i,v in color_table.iterrows() }

    # Posterior count
    f=path+"/"+pref+".posterior.count.tsv.gz"
    post = pd.read_csv(f, sep='\t')
    logging.info(f"Read posterior count from {f}")
    recol = {}
    for u in post.columns:
        v = re.match('^[A-Za-z]*_*(\d+)$', u.strip())
        if v:
            recol[v.group(0)] = v.group(1)
    if len(recol) == K:
        post.rename(columns=recol, inplace=True)
    for u in factor_header:
        post[u] = post[u].astype(float)
    post_umi = post.loc[:, factor_header].sum(axis = 0).astype(int).values
    post_weight = post.loc[:, factor_header].sum(axis = 0).values.astype(float)
    post_weight /= post_weight.sum()

    # DE genes
    f=path+"/DE/"+pref+".bulk_chisq.tsv"
    if not os.path.exists(f):
        f=path+"/"+pref+".bulk_chisq.tsv"
        if not os.path.exists(f):
            sys.exit(f"Cannot find DE file")
    de = pd.read_csv(f, sep='\t', dtype={'factor':str})
    logging.info(f"Read DE genes from {f}")
    # de.factor = de.factor.astype(int)
    top_gene = []
    top_gene_anno = []
    de['Rank'] = 0
    # Temporary: shorten unspliced gene names
    de.gene = de.gene.str.replace('unspl_', 'u_')
    # Top genes by Chi2
    de.sort_values(by=['factor','Chi2'],ascending=False,inplace=True)
    de["Rank"] = de.groupby(by = "factor").Chi2.rank(ascending=False, method = "min").astype(int)
    for k, kname in enumerate(factor_header):
        indx = de.factor.eq(kname)
        v = de.loc[indx, 'gene'].iloc[:args.n_top_gene_on_tree].values
        top_gene_anno.append(', '.join(v))
        v = de.loc[indx & ( (de.Rank < mtop) | \
                ((de.pval <= pval_max) & (de.FoldChange >= fc_min)) ), \
                'gene'].iloc[:ntop].values
        if len(v) == 0:
            top_gene.append([kname, '.'])
        else:
            top_gene.append([kname, ', '.join(v)])
    # Top genes by fold change
    de.sort_values(by=['factor','FoldChange'],ascending=False,inplace=True)
    de["Rank"] = de.groupby(by = "factor").FoldChange.rank(ascending=False, method = "min").astype(int)
    for k, kname in enumerate(factor_header):
        indx = de.factor.eq(kname)
        v = de.loc[indx & ( (de.Rank < mtop) | \
                ((de.pval <= pval_max) & (de.FoldChange >= fc_min)) ), \
                'gene'].iloc[:ntop].values
        if len(v) == 0:
            top_gene[k].append('.')
        else:
            top_gene[k].append(', '.join(v))
    # Top genes by absolute weight
    for k, kname in enumerate(factor_header):
        if post_umi[k] < 10:
            top_gene[k].append('.')
            continue
        v = post.gene.iloc[np.argsort(-post.loc[:, kname].values)[:ntop] ].values
        top_gene[k].append(', '.join(v))

    # Summary
    table = pd.DataFrame({'Factor':factor_header, 'RGB':color_table.RGB.values,
                        'Weight':post_weight, 'PostUMI':post_umi,
                        'TopGene_pval':[x[1] for x in top_gene],
                        'TopGene_fc':[x[2] for x in top_gene],
                        'TopGene_weight':[x[3] for x in top_gene] })
    oheader = ["Factor", "RGB", "Weight", "PostUMI", "TopGene_pval", "TopGene_fc", "TopGene_weight"]
    # Anchor genes used for initialization if applicable
    anchor_f = args.anchor
    if not os.path.exists(anchor_f):
        anchor_f = path + "/" + model_id + ".model_anchors.tsv"
    if os.path.exists(anchor_f):
        ak = pd.read_csv(anchor_f, sep='\t', names = ["Factor", "Anchors"], dtype={"Factor":str})
        table = table.merge(ak, on = "Factor", how = "left")
        oheader.insert(4, "Anchors")
        logging.info(f"Read anchor genes from {anchor_f}")

    table.sort_values(by = 'Weight', ascending = False, inplace=True)

    f = output_pref+".factor.info.tsv"
    table.loc[table.PostUMI.ge(10), oheader].to_csv(f, sep='\t', index=False, header=True, float_format="%.5f")

    f = output_pref+".factor.info.tsv"
    with open(f, 'r') as rf:
        lines = rf.readlines()
    header = lines[0].strip().split('\t')
    rows = [ list(enumerate(row.strip().split('\t') )) for row in lines[1:]]

    # Load template
    env = Environment(loader=FileSystemLoader(os.path.dirname(ejs)))
    template = env.get_template(os.path.basename(ejs))

    image_base64 = None
    tree_alt = None
    tree_caption = None

    if args.hc_tree:
        # Hierarchical clustering
        m = re.match("^[0\.]*(\d+)$", args.cprob_cut)
        if m is None:
            sys.exit(f"Invalid --cprob_cut, please use a number between 0 and 1 (e.g. 0.99)")
        cprob_label = m.group(1)
        cprob_cut  = float(args.cprob_cut)

        tree_f = args.tree_figure
        if not os.path.exists(args.tree_figure) or args.remake_tree:
            tree_f = os.path.dirname(color_f) + '/' + pref + ".coshc."+cprob_label+".tree.png"
            if not os.path.exists(tree_f) or args.remake_tree:
                model_f = args.model
                if not os.path.exists(args.model):
                    model_f = path + "/" + model_id + ".model_matrix.tsv.gz"
                if not os.path.exists(model_f):
                    print("Cannot find model file, will cluster based on posterior count")
                    model = post
                else:
                    model = pd.read_csv(model_f, sep='\t')
                # model_prob = np.array(model.iloc[:, 1:]).T + .1
                model_prob = np.array(model.loc[:, factor_header]).T + .1
                model_prob = model_prob / model_prob.sum(axis = 1).reshape((-1,1))

                circle = args.circle
                if not circle and args.circle_if > 0:
                    v = np.argsort(post_weight)[::-1]
                    w = np.cumsum(post_weight[v] )
                    k = np.arange(K)[w > cprob_cut][0]
                    if k > args.circle_if:
                        circle = True
                tree = visual_hc(model_prob, post_weight, top_gene_anno, \
                                node_color = node_color, factor_name = factor_header, circle = circle, \
                                output_f = tree_f, cprob_cut = cprob_cut)
        print(f"Tree figure path: {tree_f}")
        image_base64 = image_to_base64(tree_f)

        tree_alt = "Hierarchical clustering of factors based on pairwise coine distance"
        tree_caption = "Clustering of factors based on pairwise coine distance. Factors with high abundance jointly accounting for " + args.cprob_cut + " of observations are displayed."

    # Render the HTML file
    html_output = template.render(header=header, rows=rows, image_base64=image_base64, tree_image_alt=tree_alt, tree_image_caption=tree_caption)

    f=output_pref+".factor.info.html"
    with open(f, "w") as html_file:
        html_file.write(html_output)

    print(f)

if __name__ == "__main__":
    factor_report(sys.argv[1:])



# Visualize pixel level factor analysis results
# <= 5 (preferably 3) factors at a time, with specified colors
# Input file contains only top k factors and probabilities per pixel
# Meant to make use of the indexed input to plot for specified regions quickly

import sys, os, copy, re, gzip, argparse, logging, warnings
import numpy as np
import pandas as pd
from scipy.sparse import *
import cv2

from ficture.loaders.pixel_factor_loader import BlockIndexedLoader

def plot_pixel_multi(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output file full path')
    parser.add_argument('--channel_list', type=str, nargs='+', help="Select up to 3 channels/factors to visualize")
    parser.add_argument('--color_list', type=str, nargs='*', default=["144A74", "FF9900", "DD65E6", "FFEC11"], help='') # blue, orange, purple, yellow
    parser.add_argument('--color_table', type=str, default='', help='Pre-defined color map')
    parser.add_argument('--color_table_index_column', type=str, default='Name', help='')
    parser.add_argument('--input_rgb_uint8', action="store_true",help="If input rgb is from 0-255 instead of 0-1")

    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--pcut', type=float, default=0.01, help="")
    parser.add_argument('--spcut', type=float, default=0.1, help="")
    parser.add_argument('--org_coord', action='store_true', help="If the input coordinates do not include the offset (if your coordinates are from an existing figure, the offset is already factored in)")
    parser.add_argument('--full', action='store_true', help="")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Actual size (um) corresponding to each pixel in the output image")
    parser.add_argument('--categorical', action='store_true', help="Plot top factor for each pixel categorically, without probability or mixture")
    parser.add_argument('--debug', action='store_true', help="")

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None), format='%(asctime)s %(message)s', datefmt='%I:%M:%S %p')

    rgb=list("RGB")
    bgr=list("BGR") # opencv default channel order
    bgra=list("BGRA") # opencv bgra
    pcut = 0.01
    spcut = 0.1
    if len(args.channel_list) > 5:
        logging.warning("Be colorblind friendly")
    channels = args.channel_list
    # Read color table
    if os.path.exists(args.color_table):
        color_info = pd.read_csv(args.color_table, sep='\t', header=0, index_col=args.color_table_index_column)
        if color_info[rgb].max().max() > 1.1 and not args.input_rgb_uint8:
            logging.warning("Color table should contain RGB values in 0-1 range, but found values > 1, will assume unit8")
            args.input_rgb_uint8 = True
        if args.input_rgb_uint8:
            for c in rgb:
                color_info[c] = np.clip(color_info[c]/255,0,1)
        color_info.index = color_info.index.astype(str)
        channels = [x for x in channels if x in color_info.index]
        if len(channels) != len(args.channel_list):
            logging.error("Some channels are not found in the input color table")
            sys.exit(1)
        color_info = {i: np.array(color_info.loc[i, rgb]) for i in channels}
    else:
        if len(args.color_list) < len(channels):
            logging.error("Number of input colors and channels do not match")
            sys.exit(1)
        color_list = [[int(x.strip('#')[i:i+2],16)/255 for i in (0,2,4)] for x in args.color_list]
        color_info = {x:np.array(color_list[i]) for i,x in enumerate(channels)}
        ctab = pd.DataFrame(color_info, index=rgb).T
        ctab['Name'] = ctab.index
        if not args.output.endswith(".png"):
            f = args.output + ".color_table.tsv"
        else:
            f = args.output.replace(".png", ".color_table.tsv")
        ctab.to_csv(f, sep='\t', header=True, index=False)

    logging.info(f"Set up color map {color_info}")

    loader = BlockIndexedLoader(args.input, args.xmin, args.xmax, args.ymin, args.ymax, args.full, not args.org_coord)
    width = int((loader.xmax - loader.xmin + 1)/args.plot_um_per_pixel)
    height= int((loader.ymax - loader.ymin + 1)/args.plot_um_per_pixel)
    logging.info(f"Image size {height} x {width}")

    # Read input file, fill the rgb matrix
    img = np.zeros((height,width,4), dtype=np.uint8)
    img[:,:,3] = 255
    clps = {x:np.mean for x in rgb}
    clps['A'] = np.max
    for df in loader:
        if df.shape[0] == 0:
            continue
        indx = np.zeros(df.shape[0], dtype=bool)
        if args.categorical:
            indx = df.K1.isin(channels)
        else:
            for k in range(1, loader.meta['TOPK']):
                indx = indx | (df[f"K{k}"].isin(channels) & df[f"P{k}"].gt(pcut))
        df = df.loc[indx, :]
        if args.debug:
            print(df.shape[0])
        if df.shape[0] == 0:
            continue
        df['X'] = np.clip(((df.X - loader.xmin) / args.plot_um_per_pixel).astype(int),0,width-1)
        df['Y'] = np.clip(((df.Y - loader.ymin) / args.plot_um_per_pixel).astype(int),0,height-1)
        df[bgra] = 0
        if args.categorical:
            df.loc[:,rgb] = np.array(list(df.K1.map(color_info)) )
            df['A'] = 1
        else:
            for k in range(1, loader.meta['TOPK']):
                indx = df[f"K{k}"].isin(channels)
                if indx.sum() == 0:
                    continue
                df.loc[indx, rgb] += np.array(list(df.loc[indx, f"K{k}"].map(color_info))) * df.loc[indx, f"P{k}"].values.reshape((-1,1))
                df.loc[indx, 'A'] += df.loc[indx, f"P{k}"]
        if args.debug:
            print(df.shape[0], np.median(df.A), np.mean(df.A), df.A.gt(spcut).sum())
        df = df.groupby(by=['X','Y']).agg(clps).reset_index()
        df = df.loc[df.A.gt(spcut), :]
        df[bgra] = np.clip(np.around(df[bgra] * 255),0,255).astype(np.uint8)
        logging.info(f"Reading pixels... {df.X.iloc[-1]}, {df.Y.iloc[-1]}, {df.shape[0]}")
        for i,c in enumerate(bgra):
            img[df.Y.values, df.X.values, [i]*df.shape[0]] = df[c].values
        if args.debug:
            break

    if not args.output.endswith(".png"):
        args.output += ".png"
    cv2.imwrite(args.output,img)
    logging.info(f"Finished\n{args.output}")


if __name__ == "__main__":
    plot_pixel_multi(sys.argv[1:])



'''
Transform into factor space based on input model
Model contains gene names and either Dirichlet parameters (or probabilities for more general use?)
Input pixel level data will be grouped into (overlapping) hexagons
'''
import sys, os, copy, gzip, time, logging, pickle, argparse
import numpy as np
import pandas as pd
import random as rng
from sklearn.preprocessing import normalize
from sklearn.decomposition import LatentDirichletAllocation as LDA
sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from sklearn.decomposition._online_lda_fast import _dirichlet_expectation_2d

from ficture.loaders.pixel_to_unit_loader import PixelToUnit
from ficture.utils.utilt import init_latent_vars

def transform(_args):

    parser = argparse.ArgumentParser(prog = "transform")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', '--output_pref', type=str, help='')
    parser.add_argument('--model', type=str, help='')
    parser.add_argument('--feature', type=str, default='', help='')
    parser.add_argument('--key', type=str, default = 'gn', help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced')
    parser.add_argument('--major_axis', type=str, default=None, help='X or Y')
    parser.add_argument('--region_id', type=str, default=None, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=20, help='')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--thread', type=int, default=-1, help='')
    parser.add_argument('--n_move', type=int, default=3, help='')
    parser.add_argument('--hex_width', type=float, default=24, help='')
    parser.add_argument('--hex_radius', type=float, default=-1, help='')
    parser.add_argument('--precision', type=int, default=1, help='Number of digits to store spatial location (in um), 0 for integer.')
    parser.add_argument('--chunksize', type=int, default=1000000, help='Number of lines to read at a time')
    parser.add_argument('--xy_median', action='store_true', help='Output the median of pixel cooredinates inside each hexagon, default is to output hexagon centers exactly as lattice points')
    parser.add_argument('--log_norm', action='store_true', help='')
    parser.add_argument('--log_norm_size_factor', action='store_true', help='')
    parser.add_argument('--scale_const', type=float, default=-1, help='')
    parser.add_argument('--unit_sum_mean', type=float, default=-1, help='')
    parser.add_argument('--debug', type=int, default=0, help='')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    if not os.path.exists(args.input):
        sys.exit("ERROR: cannot find input file.")
    if not os.path.exists(args.model):
        sys.exit("ERROR: cannot find model file.")
    if args.hex_width <= 0 and args.hex_radius <= 0:
        sys.exit("ERROR: invalid hex_width or hex_radius")
    if not os.path.exists(os.path.dirname(args.output)):
        os.makedirs(os.path.dirname(args.output))
    logging.basicConfig(level= getattr(logging, "INFO", None))

    ### Input
    with gzip.open(args.input, 'rt') as rf:
        header=rf.readline().strip().split('\t')
    key = args.key.lower()
    header=[x.upper() if len(x) == 1 else x.lower() for x in header]
    usecol = ['X','Y','gene',key]
    adt={'X':float,'Y':float,'gene':str,args.key:int}
    if args.region_id is not None:
        usecol.append(args.region_id)
        adt[args.region_id] = str
    miss_header = [x for x in usecol if x not in header]
    if len(miss_header) > 0:
        sys.exit(f"ERROR: {miss_header} is not in the input file")
    reader = pd.read_csv(gzip.open(args.input, 'rt'), sep='\t',\
                        chunksize=args.chunksize, skiprows=1, names=header,\
                        usecols=usecol, dtype=adt)

    ### Model
    factor_header = []
    if args.model.endswith('.tsv.gz') or args.model.endswith('tsv'):
        model_mtx = pd.read_csv(args.model, sep='\t', index_col = 0)
        factor_header = list(model_mtx.columns)
        feature_kept =list(model_mtx.index)
        M, K = model_mtx.shape
        model = LDA(n_components=K, learning_method='online', batch_size=512, n_jobs = args.thread, verbose = 0)
        init_latent_vars(model, n_features = M, gamma = np.array(model_mtx).T)
    else:
        try:
            model = pickle.load(open(args.model, 'rb'))
            model.n_jobs = args.thread
            feature_kept = model.feature_names_in_
            K, M = model.components_.shape
            model.feature_names_in_ = None
            factor_header = list(np.arange(K).astype(str) )
        except:
            sys.exit("ERROR: --model must be either a tsv file containing gene name and factor-gene loadings or a pickle model object from sklearn LDA")
    if os.path.isfile(args.feature):
        feature = pd.read_csv(args.feature, sep='\t')
        kept_idx = np.where(np.in1d(feature_kept, feature.gene.values))[0]
        feature_kept = np.array(feature_kept)[kept_idx]
        model.components_ = model.components_[:, kept_idx]
        model.exp_dirichlet_component_ = np.exp(
            _dirichlet_expectation_2d(model.components_)
        )
        M = len(feature_kept)

    ft_dict = {x:i for i,x in enumerate(feature_kept)}
    logging.info(f"Model loaded with {M} features and {K} factors")

    ### Basic parameterse
    radius = args.hex_radius
    if radius < 0:
        radius = args.hex_width / np.sqrt(3)
    b_size = radius * 20
    scale_const = args.scale_const
    unit_sum_mean = args.unit_sum_mean
    if args.log_norm_size_factor:
        args.log_norm = True
    if args.log_norm:
        if hasattr(model, 'log_norm_scaling_const_'):
            scale_const = model.log_norm_scaling_const_
        if scale_const < 0:
            sys.exit("ERROR: --log_norm is specified but the normalization constant used in model fitting is unknown.")
        if args.log_norm_size_factor:
            if hasattr(model, 'unit_sum_mean_'):
                unit_sum_mean = model.unit_sum_mean_
            if unit_sum_mean < 0:
                sys.exit("ERROR: --log_norm_size_factor is specified but the size constant used in model fitting is unknown.")

    # Pixel reader
    batch_obj = PixelToUnit(reader, ft_dict, key, radius,\
                scale=1./args.mu_scale, min_ct_per_unit=args.min_ct_per_unit,\
                sliding_step=args.n_move, major_axis=args.major_axis,\
                xy_lattice=(not args.xy_median))

    # Transform
    post_count = np.zeros((K, M))
    n_unit = 0
    n_batch= 0
    oheader = ["unit",key,"x","y","topK","topP"]+[str(x) for x in range(K)]
    out_f = args.output + ".fit_result.tsv.gz"
    with gzip.open(out_f, 'wt') as wf:
        wf.write('\t'.join(oheader) + '\n')
    t0 = time.time()
    last_batch = set()
    while batch_obj.read_chunk(min_size=b_size):
        if args.log_norm_size_factor:
            rsum = batch_obj.mtx.sum(axis=1) / unit_sum_mean
            mtx = batch_obj.mtx / rsum.reshape((-1,1))
            mtx.data = np.log(mtx.data + 1) / scale_const
            mtx = mtx.tocsr()
        elif args.log_norm:
            mtx = normalize(batch_obj.mtx, norm='l1', axis=1)
            mtx.data = np.log(mtx.data + 1) / scale_const
        else:
            mtx = batch_obj.mtx
        theta = model.transform(mtx)
        post_count += np.array(theta.T @ batch_obj.mtx)
        n_batch += 1
        n_unit  += theta.shape[0]
        t1 = time.time() - t0
        batch_obj.brc['topK'] = np.argmax(theta, axis = 1)
        batch_obj.brc['topP'] = theta.max(axis = 1)
        batch_obj.brc = pd.concat([batch_obj.brc, pd.DataFrame(theta, columns=[str(x) for x in range(K)] )], axis=1)
        batch_obj.brc['x'] = batch_obj.brc.x.map(lambda x : f"{x:.{args.precision}f}")
        batch_obj.brc['y'] = batch_obj.brc.y.map(lambda x : f"{x:.{args.precision}f}")
        batch_obj.brc.rename(columns = {'hex_id':'unit'}, inplace=True)
        if len(last_batch) > 0:
            batch_obj.brc.drop(index = batch_obj.brc.index[batch_obj.brc.unit.isin(last_batch)], inplace=True)
        last_batch = set(batch_obj.brc.unit.values)
        batch_obj.brc[oheader].to_csv(out_f, sep='\t', index=False, header=False, float_format='%.3e', mode='a', compression={"method":"gzip"})
        logging.info(f"Transformed {n_batch} batches with total {n_unit} units, {t1/60:2f}min")
        if (args.debug > 0) and (n_unit >= args.debug):
            break

    out_f = args.output + ".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({'gene': feature_kept}),\
            pd.DataFrame(post_count.T, dtype='float64',\
                            columns = factor_header)],\
                axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

if __name__ == '__main__':
    transform(sys.argv[1:])



import sys, logging, importlib

def main():

    # MAYBE.. should change legacy file names to match callable module names
    module_map = {\
        "filter_by_density": "filter_poly", \
        # "filter_by_density_v1": "filter_density", \
        "filter_by_boundary": "filter_boundary", \
        "make_spatial_minibatch": "make_spatial_minibatch",\
        "make_dge": "make_dge_univ", \
        "make_sge_by_hexagon": "make_sge_by_hexagon", \
        "fit_model": "init_model_selection", \
        # "lda": "lda_univ", \
        "transform": "transform_univ", \
        "choose_color": "choose_color", \
        "plot_base": "plot_base", \
        "de_bulk": "de_bulk", \
        "factor_report": "factor_report", \
        "slda_decode": "slda_decode", \
        "plot_base": "plot_base", \
        "plot_hexagon": "plot_hexagon", \
        "plot_pixel_multi": "plot_pixel_multi", \
        "plot_pixel_full": "plot_pixel_full", \
        "plot_pixel_single": "plot_pixel_single", \
        "run_together": "run_together", \
        "init_model_from_pseudobulk": "init_from_pseudobulk", \
    }

    if len(sys.argv) < 2:
        print("Usage: ficture <command> <args>, ficture <command> -h to see arguments for each command")
        print("Available commands:\n"+"\t".join(list(module_map.keys()) ))
        return
    elif sys.argv[1] not in module_map:
        print("Unknown command: "+sys.argv[1])
        print("Available commands:\n"+"\t".join(list(module_map.keys()) ))
        return

    function_name = sys.argv[1]
    module_name = "ficture.scripts." + module_map[function_name]
    module = importlib.import_module(module_name)
    function = getattr(module, function_name)

    function(sys.argv[2:])



# Visualize pixel level factor analysis results
# Input file contains only top k factors and probabilities per pixel
# Meant to make use of the indexed input to plot for specified regions quickly

import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from scipy.sparse import *
import cv2

from ficture.loaders.pixel_factor_loader import BlockIndexedLoader

def plot_pixel_full(_args):

    parser = argparse.ArgumentParser(prog="plot_pixel_full")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output file full path')
    parser.add_argument('--category_column', type=str, default='', help='If the input contains categorical labels instead of probabilities')
    parser.add_argument('--category_map', type=str, default='', help='If needed, map the input category to a different set of names matching that specified in the color table. The mathcing could be multiple to one, input should have two columns as "From" and "To"')
    parser.add_argument('--unmapped', type=str, default='', help='')
    parser.add_argument('--color_table', type=str, help='Pre-defined color map')
    parser.add_argument('--color_table_index_column', type=str, default='Name', help='')
    parser.add_argument('--input_rgb_uint8', action="store_true",help="If input rgb is from 0-255 instead of 0-1")
    parser.add_argument('--background', type=str, default="000000", help='')

    parser.add_argument('--xmin', type=float, default=-np.inf, help="in um")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="in um")
    parser.add_argument('--xmax', type=float, default=np.inf, help="in um")
    parser.add_argument('--ymax', type=float, default=np.inf, help="in um")
    parser.add_argument('--full', action='store_true', help="Read full input")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Actual size (um) corresponding to each pixel in the output image")
    parser.add_argument('--org_coord', action='store_true', help="If the input coordinates do not include the offset (if your coordinates are from an existing figure, the offset is already factored in)")
    parser.add_argument('--plot_top', action='store_true', help="Plot top factor only")
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None), format='%(asctime)s %(message)s', datefmt='%I:%M:%S %p')

    # Read color table
    rgb=['B','G','R'] # opencv rgb order
    args.background = args.background.lstrip('#')
    match = re.search(r'^(?:[0-9a-fA-F]{3}){1,2}$', args.background)
    if not match:
        logging.warning(f"Invalid background color {args.background}")
        args.background = "000000"
    logging.info(f"Background color {args.background}")
    cdty = {x:float for x in rgb}
    cdty[args.color_table_index_column] = str
    sep=',' if args.color_table.endswith(".csv") else '\t'
    color_info = pd.read_csv(args.color_table, sep=sep, header=0, index_col=args.color_table_index_column, dtype=cdty)
    if args.input_rgb_uint8 or color_info[rgb].max().max() > 2:
        for c in rgb:
            color_info[c] = color_info[c] / 255
    # color_info.index = color_info.index.astype(str)
    logging.info(f"Read color table ({color_info.shape[0]})")
    print(color_info.index)
    dty = {}
    if args.category_column != '':
        dty[args.category_column] = str

    loader = BlockIndexedLoader(args.input, args.xmin, args.xmax, args.ymin, args.ymax, args.full, not args.org_coord, idtype=dty)
    width = int((loader.xmax - loader.xmin + 1)/args.plot_um_per_pixel)
    height= int((loader.ymax - loader.ymin + 1)/args.plot_um_per_pixel)
    logging.info(f"Image size {height} x {width}")
    print(loader.header)

    categorical = False
    if args.category_column != '':
        if args.category_column not in loader.header:
            sys.exit(f"ERROR: {args.category_column} not found in input file")
        categorical = True
        logging.info("Input is categorical")

    category_rename = False
    category_map = {}
    if os.path.exists(args.category_map):
        if args.category_map.endswith(".gz"):
            rf = gzip.open(args.category_map, "rt")
        else:
            rf = open(args.category_map, "r")
        for line in rf:
            wd = line.strip().split('\t')
            category_map[wd[0]] = wd[1]
        rf.close()
        category_rename = True
        category_list = sorted(list(set(category_map.values())))
        K = len(category_list)
        logging.info(f"Read category map ({len(category_map)}), {K} categories")
        if args.debug:
            print(category_list)

    # Read input file, fill the rgb matrix
    img = np.zeros((height,width,3), dtype=np.uint8)
    bg = args.background
    bg = [ np.uint8(int(bg[i:i+2], 16) ) for i in [0,2,4] ]
    for c in range(3):
        img[:,:,c] = bg[c]
    keptcol = ['X','Y'] + rgb
    for df in loader:
        if df.shape[0] == 0:
            continue
        df['X'] = np.clip(((df.X - loader.xmin) / args.plot_um_per_pixel).astype(int),0,width-1)
        df['Y'] = np.clip(((df.Y - loader.ymin) / args.plot_um_per_pixel).astype(int),0,height-1)
        if categorical:
            if args.debug:
                print(df.loc[~df[args.category_column].isin(category_map), :][args.category_column].unique() )
            if category_rename:
                df[args.category_column] = df[args.category_column].map(category_map)
                if args.unmapped != '':
                    df[args.category_column].fillna(args.unmapped, inplace=True)
                else:
                    df = df.loc[~df[args.category_column].isna(), :]
                if args.debug:
                    print(df.shape[0], df[args.category_column].value_counts())
            else:
                df = df.loc[df[args.category_column].isin(color_info.index), :]
            for c in rgb:
                df[c] = color_info.loc[df[args.category_column].values, c].values
            if args.debug:
                print(df[args.category_column].value_counts())
        else:
            if args.plot_top:
                for c in rgb:
                    df[c] = color_info.loc[df['K1'].values, c].values
            else:
                for c in rgb:
                    df[c] = 0
                    for k in range(1,loader.meta['TOPK']+1):
                        df[c] += color_info.loc[df['K'+str(k)].values, c].values * df['P'+str(k)].values
        df = df.groupby(by=['X','Y']).agg({c:"mean" for c in rgb}).reset_index()
        logging.info(f"Reading pixels... {df.X.iloc[-1]}, {df.Y.iloc[-1]}, {df.shape[0]}")
        for i,c in enumerate(rgb):
            df[c] = np.clip(np.around(df[c] * 255),0,255).astype(np.uint8)
            img[df.Y.values, df.X.values, [i]*df.shape[0]] = df[c].values
        if args.debug:
            break

    if not args.output.endswith(".png"):
        args.output += ".png"
    cv2.imwrite(args.output,img)
    logging.info(f"Finished\n{args.output}")

if __name__ == "__main__":
    plot_pixel_full(sys.argv[1:])



import numpy as np
import pandas as pd

### Transform between cartesian and hexagon coordinates
### "pointy-top" orientation & axial coordinate
### (https://www.redblobgames.com/grids/hexagons/)

def pixel_to_hex(pts, size, offset_x = 0, offset_y = 0):
    n,d = pts.shape
    assert d == 2
    mtx = np.array([[np.sqrt(3)/3,-1/3],[0,2/3]])
    hex_frac = mtx @ pts.transpose()
    hex_frac /= size
    hex_frac[0,] += offset_x
    hex_frac[1,] += offset_y
    rx = np.round(hex_frac[0,])
    ry = np.round(hex_frac[1,])
    hex_frac[0,] -= rx
    hex_frac[1,] -= ry
    indx = np.abs(hex_frac[0,]) < np.abs(hex_frac[1,])
    x = rx+np.round(hex_frac[0,]+0.5*hex_frac[1,])
    y = ry
    x[indx] = rx[indx]
    y[indx] = ry[indx]+np.round(hex_frac[1,indx]+0.5*hex_frac[0,indx])
    return x.astype(int), y.astype(int)

def hex_to_pixel(x,y,size,offset_x=0,offset_y=0):
    if hasattr(x, "__len__"):
        assert hasattr(y, "__len__")
        x = np.array(x)
        y = np.array(y)
        assert len(y) == len(x)
    ptx = size * (np.sqrt(3) * (x-offset_x) + np.sqrt(3)/2 * (y-offset_y))
    pty = size * 3/2 * (y-offset_y)
    return ptx, pty

def collapse_to_hex(df, hex_width = -1, radius = -1, n_move = 1, key = "Count", ):
    if hex_width > 0:
        radius = hex_width / np.sqrt(3)
    assert radius > 0
    df.rename(columns={"x":"X","y":"Y"}, inplace=True)
    assert "X" in df.columns and "Y" in df.columns
    brc = pd.DataFrame()
    for i in range(n_move):
        for j in range(n_move):
            cnt = pd.DataFrame()
            cnt["hex_x"], cnt["hex_y"] = pixel_to_hex(df.loc[:, ['X','Y']].values, radius, i/n_move, j/n_move)
            cnt[key] = df[key].values
            cnt = cnt.groupby(by = ['hex_x','hex_y']).agg({key:sum}).reset_index()
            cnt["offset"] = f"{i}_{j}"
            cnt["ID"] = list(zip(cnt.hex_x, cnt.hex_y, cnt.offset))
            cnt["x"], cnt["y"] = hex_to_pixel(cnt.hex_x, cnt.hex_y, radius, i/n_move, j/n_move)
            cnt.drop(columns=["hex_x","hex_y","offset"], inplace=True)
            brc = pd.concat([brc, cnt.reset_index(drop=True)])
    return brc



import sys, os, gzip, gc, argparse, warnings, logging
import numpy as np
import pandas as pd
import sklearn.neighbors
import sklearn.mixture

from ficture.utils.filter_fn import filter_by_density_mixture

def filter_by_density_v1(_args):

    parser = argparse.ArgumentParser(prog = "filter_by_density")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--ref_pt', type=str, help='')
    parser.add_argument('--filter_based_on', type=str, default="Count", help='')
    parser.add_argument('--feature', type=str, default='', help='')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--filter_batch_size', type=float, default=5000, help='In um, along the streaming axis')
    parser.add_argument('--filter_batch_ncut', type=int, default=1, help='')
    parser.add_argument('--major_axis', type=str, default="X", help='')
    parser.add_argument('--precision_um', type=float, default=2, help='')
    parser.add_argument('--log', default = '', type=str, help='files to write log to')
    parser.add_argument('--max_npts_to_fit_model', type=float, default=1e6, help='')
    parser.add_argument('--min_abs_mol_density_squm_dense', type=float, default=0.2, help='Lowerbound for dense tissue region')
    parser.add_argument('--min_abs_mol_density_squm', type=float, default=0.02, help='A safe lowerbound to remove very sparse technical noise')
    parser.add_argument('--hard_threshold', type=float, default=-1, help='If provided, filter by hard threshold (number of molecules per squared um)')
    parser.add_argument('--radius', type=float, default=7, help='')
    parser.add_argument('--hex_n_move', type=int, default=6, help='')

    parser.add_argument('--xmin', type=float, default=-1, help='In um')
    parser.add_argument('--xmax', type=float, default=np.inf, help='In um')
    parser.add_argument('--ymin', type=float, default=-1, help='In um')
    parser.add_argument('--ymax', type=float, default=np.inf, help='In um')

    parser.add_argument('--redo_filter', action='store_true')
    parser.add_argument('--anchor_only', action='store_true')
    parser.add_argument('--debug', action='store_true')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    if os.path.exists(args.output):
        warnings.warn("Output file already exists")
    if not os.path.exists(args.input):
        sys.exit("Cannot find input file")
    major_axis = args.major_axis.upper()
    if major_axis not in ["X", "Y"]:
        sys.exit("Invalid --major_axis")
    minor_axis = "X" if major_axis == "Y" else "Y"
    if args.xmax <= 0:
        args.xmax = np.inf
    if args.ymax <= 0:
        args.ymax = np.inf

    if args.log != '':
        try:
            logging.basicConfig(filename=args.log, filemode='a', encoding='utf-8', level=logging.INFO)
        except:
            logging.basicConfig(level= getattr(logging, "INFO", None))
    else:
        logging.basicConfig(level= getattr(logging, "INFO", None))

    key      = args.filter_based_on
    mu_scale = 1./args.mu_scale
    n_move   = args.hex_n_move
    radius   = args.radius
    hex_area = radius**2*3*np.sqrt(3)/2
    chunk_size = 1000000

    gene_kept = set()
    if os.path.exists(args.feature):
        feature = pd.read_csv(args.feature, sep='\t', header=0)
        gene_kept = set(feature.gene.values)

    if os.path.exists(args.ref_pt) and not args.redo_filter:
        pt = pd.read_csv(args.ref_pt, sep='\t')
        logging.info(f"Read existing anchor points")
    else:
        pt = pd.DataFrame()
        df=pd.DataFrame()
        for chunk in pd.read_csv(gzip.open(args.input, 'rb'),\
            sep='\t', usecols=["Y","X","gene",key], chunksize=chunk_size):
            full = chunk.shape[0] == chunk_size
            if len(gene_kept) != 0:
                chunk = chunk.loc[chunk.gene.isin(gene_kept)]
            chunk['X'] = chunk.X.values * mu_scale
            chunk['Y'] = chunk.Y.values * mu_scale
            chunk = chunk.loc[(chunk.X >= args.xmin) & (chunk.X <= args.xmax) &\
                            (chunk.Y >= args.ymin) & (chunk.Y <= args.ymax)]
            if chunk.shape[0] == 0:
                continue
            if args.precision_um > 0:
                chunk['X'] = np.around(chunk.X.values/args.precision_um,0).astype(int)*args.precision_um
                chunk['Y'] = np.around(chunk.Y.values/args.precision_um,0).astype(int)*args.precision_um
            chunk = chunk.groupby(by = ["X", "Y"]).agg({key:sum}).reset_index()
            df = pd.concat([df, chunk])
            hst, hed = df[major_axis].iloc[0], df[major_axis].iloc[-1]
            wst, wed = df[minor_axis].min(), df[minor_axis].max()
            logging.info(f"Current chunk size {hed-hst} x {wed-wst}, collapsed into {df.shape[0]} pts")

            if hed - hst < args.filter_batch_size * .9 and full:
                continue

            st = hst
            ed = st + args.filter_batch_size
            wstep = (wed - wst + 1) / args.filter_batch_ncut
            df['win'] = ((df[minor_axis] - wst)/wstep).astype(int).astype(str)
            ### Detect grid points falling inside dense tissue region
            for w in df.win.unique():
                indx = df.win.eq(w)
                xmin, ymin = df.loc[indx, ['X','Y']].min()
                xmax, ymax = df.loc[indx, ['X','Y']].max()
                sub, m0, m1 = filter_by_density_mixture(df.loc[indx, :], key, radius, n_move, args)
                pt = pd.concat([pt, sub])
                logging.info(f"Window {str(w)} ({xmax-xmin:.1f} X {ymax-ymin:.1f} ):\t{m0:.3f} v.s. {m1:.3f}")
            df=pd.DataFrame()
            if args.debug:
                break

        if not args.debug:
            pt.x = np.around(np.clip(pt.x.values,0,np.inf)/args.precision_um,0).astype(int)
            pt.y = np.around(np.clip(pt.y.values,0,np.inf)/args.precision_um,0).astype(int)
            pt.drop_duplicates(inplace=True)
            pt.x = pt.x * args.precision_um
            pt.y = pt.y * args.precision_um
            pt.to_csv(args.ref_pt, sep='\t', index=False, header=True)

        logging.info(f"Finished identifying anchor points in tissue region.")

    if args.anchor_only:
        sys.exit()


    ref=sklearn.neighbors.BallTree(np.array(pt.loc[:, ['x','y']]))
    with gzip.open(args.input, 'rt') as rf:
        header = rf.readline()
    if args.output.endswith(".gz"):
        with gzip.open(args.output, 'wt') as wf:
            _=wf.write(header)
    else:
        with open(args.output, 'w') as wf:
            _=wf.write(header)

    for chunk in pd.read_csv(gzip.open(args.input, 'rb'),\
        sep='\t', header=0, chunksize=chunk_size):
        if len(gene_kept) != 0:
            chunk = chunk.loc[chunk.gene.isin(gene_kept)]
            if chunk.shape[0] == 0:
                continue
        x = chunk.X.values * mu_scale
        y = chunk.Y.values * mu_scale
        chunk = chunk.loc[ (x >= args.xmin) & (x <= args.xmax) &\
                        (y >= args.ymin) & (y <= args.ymax) ]
        if chunk.shape[0] == 0:
            continue
        dv, iv = ref.query(X=np.array(chunk.loc[:, ["X","Y"]]) * mu_scale, \
                        k=1, return_distance=True, sort_results=False)
        dv = dv.squeeze()
        chunk = chunk.loc[dv < radius, :]
        if chunk.shape[0] == 0:
            continue
        logging.info(f"Output {chunk.shape[0]} rows ...")
        chunk.to_csv(args.output, mode='a', sep='\t', index=False, header=False)
        if args.debug:
            break

if __name__ == "__main__":
    filter_by_density_v1(sys.argv[1:])



# Functions for fitting augmented Latent Dirichlet Allocation
# with online variational Bayes (VB).

import sys, io, os, re, time, copy, subprocess, logging

import numpy as np
from sklearn.utils import check_random_state
from scipy.special import gammaln, psi, logsumexp, expit, logit
from scipy.sparse import *
from sklearn.preprocessing import normalize
from sklearn.decomposition._online_lda_fast import (
    _dirichlet_expectation_1d, _dirichlet_expectation_2d,
)

class OnlineLDA:
    """
    Implements online VB for LDA as described in (Hoffman et al. 2010).
    """
    def __init__(self, vocab, K, N, alpha = None, eta = None, tau0=9, kappa=.7, zeta = 0, iter_inner = 50, tol = 1e-4, iter_gamma = 10, verbose = 0, seed = None):
        """
        Arguments:
        K: Number of topics
        vocab: A list of features
        D:     Total number of pixels in the entire dataset.
        alpha: Hyperparameter for the prior on weight vectors theta
        eta:   Hyperparameter for prior on topics beta
        tau0:  A (non-negative) learning parameter that downweights earl iterations
        kappa: Learning rate (exponential decay), should be between
               (0.5, 1.0] to guarantee asymptotic convergence.
        --- Experimental ---
        zeta:  Weight of the proximal contamination penalty
        iter_gamma: Maximum number of iterations when maximizing the "penalized ELBO" w.r.t. gamma (there is no analytical solution)
        """
        self._vocab = vocab
        self._K = K
        self._M = len(self._vocab)
        self._N = N
        self._tau0 = tau0 + 1
        self._kappa = kappa
        self._zeta = zeta
        self._updatect = 0
        self._max_iter_inner = iter_inner
        self._max_iter_gamma = iter_gamma
        self._tol = tol
        self._verbose = verbose
        self._Elog_beta = None      # K x M
        self._lambda = None         # K x M
        self.rng_ = check_random_state(seed)

        self._alpha = alpha # 1D array
        if self._alpha is None:
            self._alpha = np.ones(self._K)/self._K
        elif np.isscalar(self._alpha):
            self._alpha = np.ones(self._K)*self._alpha
        else:
            self._alpha = self._alpha.reshape(-1)
            assert self._alpha.shape[0] == self._K, "Invalid alpha"

        self._eta = eta     # 1D array
        if self._eta is None:
            self._eta = (np.ones(self._M)/self._K).reshape((1,-1))
        elif np.isscalar(self._eta):
            self._eta = (np.ones(self._M)*self._eta).reshape((1,-1))
        else:
            self._eta = self._eta.reshape((1, -1))
            assert self._eta.shape[1] == self._M, "Invalid eta"

    def init_global_parameter(self, m_lambda=None):
        # Initialize the variational distribution q(beta|lambda)
        if m_lambda is None:
            self._lambda = self.rng_.gamma(100., 1./100., (self._K, self._M))
        else:
            self._lambda = m_lambda
        self._Elog_beta = _dirichlet_expectation_2d(self._lambda)

    def do_e_step(self, batch):
        """
        Update local parameters, compute sufficient stats for M step
        """
        Xb = batch.mtx @ self._Elog_beta.T         # Dense, N x K
        if issparse(Xb):
            Xb = Xb.toarray()
        c_indx, r_indx = batch.psi.nonzero()
        # Initialize the variational distribution q(theta|gamma)
        if batch.alpha is None:
            batch.alpha = np.broadcast_to(self._alpha, (batch.n, self._K))
        if batch.gamma is None:
            batch.gamma = copy.copy(batch.alpha)
        gamma_old = copy.copy(batch.gamma)
        Elog_theta = _dirichlet_expectation_2d(batch.gamma) # n x K
        meanchange = self._tol + 1
        it = 0

        while it < self._max_iter_inner and meanchange > self._tol:

            batch.phi = batch.psi @ Elog_theta + Xb
            batch.phi = np.exp(batch.phi -\
                               logsumexp(batch.phi, axis = 1).reshape((-1, 1)) )
            psi_hat = batch.phi[c_indx, 0] * Elog_theta[r_indx, 0] # \sum_k phi_ik Elog[theta_jk]
            for k in range(1, self._K):
                psi_hat += batch.phi[c_indx, k] * Elog_theta[r_indx, k]
            psi_hat += np.multiply(Xb, batch.phi).sum(axis = 1).reshape(-1)[c_indx] # \sum_k phi_ik log Pik
            batch.psi.data = psi_hat
            batch.psi += batch.ElogO
            batch.psi.data = np.exp(batch.psi.data)
            batch.psi = normalize(batch.psi, norm='l1', axis=1)
            batch.gamma = batch.alpha + batch.psi.T @ batch.phi
            Elog_theta = _dirichlet_expectation_2d(batch.gamma)

            meanchange = np.abs(batch.gamma - gamma_old).max(axis=1).mean()
            gamma_old = copy.copy(batch.gamma)
            it += 1
            # if self._verbose > 2: # debug
            #     print( np.around([batch.psi.sum(axis = 1).min(),\
            #                       batch.phi.sum(axis = 1).min()], 2) )
            if self._verbose > 2 or (self._verbose > 1 and it % 10 == 0):
                logging.info(f"E-step, update phi, psi, gamma: {it}-th iteration, mean change {meanchange:.4f}")

        sstats = batch.phi.T @ batch.mtx # K x M
        batch.ll = batch.psi.T @ batch.mtx @ self._Elog_beta.T
        ll_norm = logsumexp(batch.ll, axis = 1)
        ll_tot = 0
        for k in range(self._K):
            ll_tot += (batch.ll[:, k] - ll_norm).sum()
        batch.ll = ll_tot / batch.n
        return sstats

    def update_lambda_penalized(self, batch):
        assert self._zeta > 0 and self._zeta < 1, "zeta must be in (0, 1) for penalized update"
        assert batch.anchor_adj is not None, "batch.anchor_adj must be provided for penalized update"
        if issparse(batch.anchor_adj):
            assert (batch.anchor_adj.diagonal() > 0).sum() >= batch.n, "anchor_adj must contain self-loops"
        else:
            assert (np.diag(batch.anchor_adj) > 0).sum() >= batch.n, "anchor_adj must contain self-loops"
        # E step to update gamma, phi | lambda for mini-batch
        lambda_org = self._eta + self.do_e_step(batch) # K X M
        theta = normalize(batch.gamma, norm='l1', axis=1) # n x K, E[theta]
        ckl = theta.T @ normalize(batch.anchor_adj, norm='l1', axis=1) @ theta
        ckl /= theta.sum(axis = 0).reshape((-1, 1)) # K x K, factor spatial proximity
        np.fill_diagonal(ckl, 0)
        lam_sum = lambda_org.sum(axis = 1)
        it = 0
        meanchange = self._tol + 1
        while it < self._max_iter_gamma and meanchange > self._tol:
            lambda_new = np.zeros((self._K, self._M))
            for k in range(self._K):
                avg = ckl[[k], :] @ (normalize(lambda_org, norm='l1', axis=1)*lam_sum[k])
                adj = np.clip(lambda_org[k, :] - self._zeta * avg, 0, None)
                lambda_new[k, :] = adj / adj.sum() * lam_sum[k]
            v = np.abs(lambda_new - lambda_org).max(axis = 1) / lambda_org.sum(axis = 1)
            meanchange = v.mean()
            if self._verbose > 2 or (self._verbose > 1 and it % 10 == 0) or (self._verbose > 1 and it == self._max_iter_gamma - 1):
                logging.info(f"Penalized M-step, update lambda : {it}-th iteration, mean max change {meanchange:.4f}")
            lambda_org = copy.copy(lambda_new)
            it += 1

        if self._verbose > 0:
            # Estimate likelihood for current values of lambda.
            scores = self.approx_score(batch)
            logging.info(f"{self._updatect}-th global update. Scores: " + ", ".join(['%.2e'%x for x in scores]))

        rhot = pow(self._tau0 + self._updatect, -self._kappa)
        self._lambda = (1-rhot) * self._lambda + \
                       rhot * ((self._N / batch.N) * lambda_org)
        self._Elog_beta = _dirichlet_expectation_2d(self._lambda)
        self._updatect += 1
        return 1

    def update_lambda(self, batch):
        """
        Process one minibatch
        """
        # rhot will be between 0 and 1, and says how much to weight
        # the information we got from this mini-batch.

        # E step to update gamma, phi | lambda for mini-batch
        sstats = self.do_e_step(batch)
        # Estimate likelihood for current values of lambda.
        scores = self.approx_score(batch)
        if self._verbose > 0:
            logging.info(f"{self._updatect}-th global update. Scores: " + ", ".join(['%.2e'%x for x in scores]))

        rhot = pow(self._tau0 + self._updatect, -self._kappa)
        self._lambda = (1-rhot) * self._lambda + \
                       rhot * ((self._N / batch.N) * (self._eta + sstats) )
        self._Elog_beta = _dirichlet_expectation_2d(self._lambda)
        self._updatect += 1
        return scores

    def approx_score(self, batch):

        score_gamma = 0
        score_beta  = 0
        Elog_theta = _dirichlet_expectation_2d(batch.gamma)

        # E[log p(x | theta, phi, beta)]
        score_pixel = batch.mtx @ self._Elog_beta.T
        score_pixel = np.multiply(batch.phi, score_pixel/batch.N).sum()
        score_patch = batch.ll

        # E[log p(theta | alpha) - log q(theta | gamma)]
        score_gamma += np.sum((batch.alpha - batch.gamma)*Elog_theta)
        score_gamma += np.sum(gammaln(batch.gamma)) - np.sum(gammaln(np.sum(batch.gamma, 1)))
        score_gamma /= batch.n

        # E[log p(beta | eta) - log q (beta | lambda)]
        score_beta += np.sum((self._eta-self._lambda)*self._Elog_beta)
        score_beta += np.sum(gammaln(self._lambda)) - np.sum(gammaln(np.sum(self._lambda, 1)))

        return((score_pixel, score_patch, score_gamma, score_beta, self._N * (score_patch+score_gamma) + score_beta))

    def coherence_pmi(self, pw, pseudo_ct = 200, top_gene_n = 100):
        """
        Topic coherence (pointwise mutual information)
        """
        topic_pmi = []
        top_gene_n = np.min([top_gene_n, self._M])
        pw = pw / pw.sum()
        for k in range(self._K):
            indx = np.argsort(-self._Elog_beta[k, :])[:top_gene_n]
            b = np.exp(self._Elog_beta[k, indx] )
            b = np.clip(b, 1e-6, 1.-1e-6)
            w = 1. - np.power(1.-pw[indx], pseudo_ct)
            w = w.reshape((-1, 1)) @ w.reshape((1, -1))
            p0 = 1.-np.power(1-b, pseudo_ct)
            p0 = p0.reshape((-1, 1)) @ p0.reshape((1, -1))
            pmi = np.log(p0) - np.log(w)
            np.fill_diagonal(pmi, 0)
            pmi = np.round(pmi.mean(), 3)
            topic_pmi.append(pmi)
        return topic_pmi



import sys, io, os, gzip, glob, copy, re, time, warnings, pickle, argparse, logging
from collections import defaultdict,Counter
import numpy as np
import pandas as pd
from datetime import datetime

from scipy.sparse import *
from sklearn.utils import check_random_state
from sklearn.preprocessing import normalize
from joblib import Parallel, delayed
from sklearn.decomposition import LatentDirichletAllocation as LDA

sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from ficture.utils.utilt import gen_even_slices, chisq, make_mtx_from_dge
from ficture.loaders.unit_loader import UnitLoader

def fit_model(_args):

    parser = argparse.ArgumentParser()
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='')
    parser.add_argument('--nFactor', type=int, help='')
    parser.add_argument('--feature', type=str, default = '', help='')
    parser.add_argument('--key', type=str, default='gn', help='')
    parser.add_argument('--R', type=int, default=5, help='')
    parser.add_argument('--epoch_init', type=int, default=1, help='')
    parser.add_argument('--epoch', type=int, default=1, help='')
    parser.add_argument('--test_split', type=float, default=.5, help='')
    parser.add_argument('--thread', type=int, default=1, help='')
    parser.add_argument('--seed', type=int, default=-1, help='')

    parser.add_argument('--log_norm', action='store_true', help='')
    parser.add_argument('--log_norm_size_factor', action='store_true', help='')
    parser.add_argument('--alpha', type=float, default=1, help='')
    parser.add_argument('--tau', type=int, default=9, help='')
    parser.add_argument('--kappa', type=float, default=0.7, help='')

    parser.add_argument('--unit_label', default = 'random_index', type=str, help='Which column to use as unit identifier')
    parser.add_argument('--feature_label', default = "gene", type=str, help='Which column to use as feature identifier')
    parser.add_argument('--unit_attr', type=str, nargs='+', default=[], help='')
    parser.add_argument('--epoch_id_length', type=int, default=2, help='')
    parser.add_argument('--min_ct_per_unit', type=int, default=50, help='')
    parser.add_argument('--min_ct_per_feature', type=int, default=50, help='')
    parser.add_argument('--debug', action='store_true', help='')

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    seed = int(args.seed)
    if seed <= 0:
        seed = int(datetime.now().timestamp()) % 2147483648
    rng = check_random_state(seed)

    key = args.key
    thread = args.thread
    R = args.R
    K = args.nFactor
    b_size = 512
    topM = 10
    score_feature_min = 100
    unit_key = args.unit_label.lower()
    unit_attr = [x.lower() for x in args.unit_attr]
    gene_key = args.feature_label.lower()
    logging.basicConfig(level= getattr(logging, "INFO", None))

    feature_list = None
    if os.path.isfile(args.feature):
        feature_list = pd.read_csv(args.feature, sep='\t', dtype={gene_key:str})[gene_key].tolist()
        feature_list = list(set(feature_list))

    feature, brc, mtx_org, ft_dict, bc_dict = make_mtx_from_dge(args.input,\
        min_ct_per_unit = args.min_ct_per_unit, \
        min_ct_per_feature = args.min_ct_per_feature, \
        feature_list = feature_list, \
        unit = args.unit_label, key = key)
    unit_sum = mtx_org.sum(axis = 1)
    unit_sum_mean = np.mean(unit_sum)
    size_factor = unit_sum / unit_sum_mean
    gene_f = feature.Weight.values

    N, M = mtx_org.shape
    logging.info(f"Read data with {N} units, {M} features")
    Ntrain = int(N*args.test_split)
    Ntest = N - Ntrain
    train_idx = set(rng.choice(N, Ntrain, replace=False) )
    test_idx = set(range(N)) - train_idx
    train_idx = sorted(list(train_idx))
    test_idx = sorted(list(test_idx))

    scale_const = 1
    if args.log_norm_size_factor:
        mtx_log_norm = mtx_org / size_factor.reshape((-1, 1))
        mtx_log_norm.data = np.log(mtx_log_norm.data + 1)
        scale_const = np.log(1/np.quantile(size_factor, q=.95) + 1)
        mtx_log_norm.data = mtx_log_norm.data / scale_const
    elif args.log_norm:
        mtx_log_norm = normalize(mtx_org, norm = 'l1', axis = 1)
        mtx_log_norm.data = np.log(mtx_log_norm.data + 1)
        scale_const = np.log(1/np.quantile(unit_sum, q=.95) + 1)
        mtx_log_norm.data = mtx_log_norm.data / scale_const
    else:
        mtx_log_norm = mtx_org

    mtx_log_norm = mtx_log_norm.tocsr()
    results = {}
    coh_score = []
    mtx = mtx_org[test_idx, :].tocsc()
    factor_header = list(np.arange(K).astype(str) )
    for r in range(R):
        t0 = time.time()
        model = LDA(n_components=K, learning_method='online', batch_size=b_size, total_samples = N, learning_offset = args.tau, learning_decay = args.kappa, doc_topic_prior = args.alpha, n_jobs = thread, verbose = 0, random_state=seed)
        for e in range(args.epoch_init):
            rng.shuffle(train_idx)
            _ = model.partial_fit(mtx_log_norm[train_idx, :])
        score_train = model.score(mtx_log_norm[train_idx, :])/Ntrain
        score_test = model.score(mtx_log_norm[test_idx, :])/Ntest
        logging.info(f"{r}: {score_train:.2f}, {score_test:.2f}")
        # Transform the test set
        theta = model.transform(mtx_log_norm[test_idx, :])
        topk = theta.argmax(axis = 1)
        logging.info(f"{Counter(topk)}")
        # Get DE genes from the test data
        info = mtx.T @ theta
        info = pd.DataFrame(info, columns = factor_header)
        info.index = feature.gene.values
        info['gene_tot'] = info[factor_header].sum(axis = 1)
        info.drop(index = info.index[info.gene_tot < score_feature_min], inplace = True)
        logging.info(f"Computing naive DE for {len(info)} genes")
        total_k = np.array(info[factor_header].sum(axis = 0) )
        total_umi = info[factor_header].sum().sum()
        res = []
        for k, kname in enumerate(factor_header):
            idx_slices = [idx for idx in gen_even_slices(len(info), thread)]
            with Parallel(n_jobs=thread, verbose=0) as parallel:
                result = parallel(delayed(chisq)(kname, \
                            info.iloc[idx, :].loc[:, [kname, 'gene_tot']],\
                            total_k[k], total_umi) for idx in idx_slices)
            res += [item for sublist in result for item in sublist]
        chidf=pd.DataFrame(res,columns=['gene','factor','Chi2','pval','FoldChange','gene_total'])
        chidf["Rank"] = chidf.groupby(by = "factor")["Chi2"].rank(ascending=False)
        chidf.gene_total = chidf.gene_total.astype(int)
        chidf.sort_values(by=['factor','Chi2'],ascending=[True,False],inplace=True)
        # Compute a "coherence" score using top DE gene co-occurrence
        score = []
        for k in range(K):
            topm = min(topM, chidf.factor.eq(str(k)).sum())
            if topm < topM:
                logging.info(f"Factor {k} has only {topm} over-expressed genes")
            if topm == 0:
                continue
            wd_idx = chidf.loc[chidf.factor.eq(str(k))].gene.iloc[:topm].map(ft_dict).values
            wd_idx = sorted( list(wd_idx), key = lambda x : -gene_f[x])
            s = 0
            for ii in range(topm - 1):
                for jj in range(ii+1, topm):
                    i = wd_idx[ii]
                    j = wd_idx[jj]
                    idx = mtx.indices[mtx.indptr[i]:mtx.indptr[i+1]]
                    denom = mtx[:, [i]].toarray()[idx] * gene_f[j] / gene_f[i]
                    num = mtx[:, [j]].toarray()[idx]
                    s += (theta[idx, k].reshape((-1, 1)) * np.log(num/denom + 1)).sum()
            s0 = s / theta[:, k].sum()
            coh_score.append([r, k, s, s0])
            score.append(s0)

        t1 = time.time() - t0
        t0 = time.time()
        logging.info(f"R={r}, {np.mean(score):.2f}, {np.median(score):.2f}, {t1:.2f}s")
        results[r] = {'score_train':score_train, 'score_test':score_test, 'model':model, 'coherence':score}

    pickle.dump(results, open(args.output + ".model_selection_candidates.p", 'wb'))
    coh_score = pd.DataFrame(coh_score, columns = ["R","K","Score0","Score"])
    coh_score.to_csv(args.output + ".coherence.tsv", sep='\t', index = False)
    v = coh_score.groupby(by = "R").Score.mean()
    v = v.sort_values(ascending = False)
    print(v)

    ### Further update the selected model
    best_r = v.index[0]
    model = results[best_r]['model']
    epoch = args.epoch_init * (1 - args.test_split)
    n_unit = 0
    chunksize = 2000000
    with gzip.open(args.input, 'rt') as rf:
        header = rf.readline().strip().split('\t')
    header = [x.lower() if x != key else x for x in header]
    adt = {unit_key:str, key:int}
    adt.update({x:str for x in unit_attr})
    while epoch < args.epoch:
        reader = pd.read_csv(gzip.open(args.input, 'rt'), \
                sep='\t',chunksize=chunksize, skiprows=1, names = header,
                usecols=[unit_key,gene_key,key], dtype=adt)
        batch_obj =  UnitLoader(reader, ft_dict, key, \
            batch_id_prefix=args.epoch_id_length, \
            min_ct_per_unit=args.min_ct_per_unit,
            unit_id=unit_key,unit_attr=[])
        while batch_obj.update_batch(b_size):
            N = batch_obj.mtx.shape[0]
            if args.log_norm_size_factor:
                rsum = batch_obj.mtx.sum(axis = 1) / unit_sum_mean
                mtx_fit = batch_obj.mtx / rsum.reshape((-1, 1))
                mtx_fit.data = np.log(mtx_fit.data + 1) / scale_const
            elif args.log_norm:
                mtx_fit = normalize(batch_obj.mtx, norm='l1', axis=1)
                mtx_fit.data = np.log(mtx_fit.data + 1) / scale_const
            else:
                mtx_fit = batch_obj.mtx
            _ = model.partial_fit(mtx_fit)
            n_unit += N
            if args.debug:
                logl = model.score(mtx_fit) / N
                e = len(batch_obj.batch_id_list)
                logging.info(f"Epoch {e-1}, finished {n_unit} units. batch logl: {logl:.4f}")
            if len(batch_obj.batch_id_list) > args.epoch:
                break
        if args.epoch_id_length > 0:
            epoch += len(batch_obj.batch_id_list)
        else:
            epoch += 1

    # Relabel factors
    weight = model.components_.sum(axis=1)
    ordered_k = np.argsort(weight)[::-1]
    model.components_ = model.components_[ordered_k,:]
    model.exp_dirichlet_component_ = model.exp_dirichlet_component_[ordered_k,:]

    ### Rerun all units once and store results
    oheader = ["unit",key,"x","y","topK","topP"]+factor_header
    dtp = {'topK':int,key:int,"unit":str}
    dtp.update({x:float for x in ['topP']+factor_header})
    res_f = args.output+".fit_result.tsv.gz"
    logging.info(f"Result file {res_f}")
    theta = model.transform(mtx_log_norm)
    post_count = np.array(theta.T @ mtx_org)
    brc.rename(columns = {'j':'unit', 'X':'x', 'Y':'y'}, inplace = True)
    brc['topK'] = np.argmax(theta, axis = 1).astype(int)
    brc['topP'] = np.max(theta, axis = 1)
    brc = pd.concat((brc, pd.DataFrame(theta, columns = factor_header )), axis = 1)
    brc = brc.astype(dtp)
    brc[oheader].to_csv(res_f, sep='\t', float_format="%.4e", index=False, header=True, compression={"method":"gzip"})

    out_f = args.output+".posterior.count.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature.gene.values}),\
            pd.DataFrame(post_count.T, columns = factor_header, dtype='float64')],\
        axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.2f', compression={"method":"gzip"})

    model.feature_names_in_ = feature.gene.values
    model.log_norm_scaling_const_ = scale_const
    model.unit_sum_mean_ = np.mean(unit_sum)
    pickle.dump(model, open(args.output + ".model.p", 'wb'))

    out_f = args.output + ".model_matrix.tsv.gz"
    pd.concat([pd.DataFrame({gene_key: feature.gene.values }),\
            pd.DataFrame(model.components_.T, columns = factor_header, dtype='float64')], \
        axis = 1).to_csv(out_f, sep='\t', index=False, float_format='%.4e', compression={"method":"gzip"})

if __name__ == "__main__":
    fit_model(sys.argv[1:])



import sys, os, copy, gc, re, gzip, pickle, argparse, logging, warnings
import numpy as np
import pandas as pd
from random import shuffle

import matplotlib as mpl
import matplotlib.pyplot as plt
from PIL import Image

from scipy.sparse import *
import sklearn.neighbors
import sklearn.preprocessing

from ficture.utils.hexagon_fn import *
from ficture.utils.utilt import plot_colortable

def plot_base(_args):

    parser = argparse.ArgumentParser(prog = "plot_base")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--output', type=str, help='Output prefix')
    parser.add_argument('--fill_range', type=float, default=0, help="um")
    parser.add_argument('--batch_size', type=float, default=500, help="")
    parser.add_argument("--tif", action='store_true', help="Store as 16-bit tif instead of png")
    parser.add_argument('--scale', type=float, default=-1, help="")
    parser.add_argument('--origin', type=int, default=[0,0], help="{0, 1} x {0, 1}, specify how to orient the image w.r.t. the coordinates. (0, 0) means the lower left corner has the minimum x-value and the minimum y-value; (0, 1) means the lower left corner has the minimum x-value and the maximum y-value;")
    parser.add_argument('--category_column', type=str, default='', help='')

    parser.add_argument('--color_table_category_name', type=str, default='Name', help='When --category_column is provided, which column to use as the category name')
    parser.add_argument('--binary_cmap_name', type=str, default="plasma", help="Name of Matplotlib colormap to use for ploting individual factors")
    parser.add_argument('--color_table', type=str, default='', help='Pre-defined color map')
    parser.add_argument('--input_rgb_uint8', action="store_true",help="If input rgb is from 0-255 instead of 0-1")
    parser.add_argument('--cmap_name', type=str, default="turbo", help="Name of Matplotlib colormap, only used when --color_table is not provided")

    parser.add_argument("--plot_fit", action='store_true', help="")
    parser.add_argument('--xmin', type=float, default=-np.inf, help="")
    parser.add_argument('--ymin', type=float, default=-np.inf, help="")
    parser.add_argument('--xmax', type=float, default=np.inf, help="")
    parser.add_argument('--ymax', type=float, default=np.inf, help="")
    parser.add_argument('--plot_um_per_pixel', type=float, default=1, help="Size of the output pixels in um")

    parser.add_argument("--skip_mixture_plot", action='store_true', help="")
    parser.add_argument("--plot_discretized", action='store_true', help="")
    parser.add_argument("--plot_individual_factor", action='store_true', help="")
    parser.add_argument("--debug", action='store_true', help="")

    args = parser.parse_args(_args)
    if len(_args) == 0:
        parser.print_help()
        return

    logging.basicConfig(level= getattr(logging, "INFO", None))
    dt = np.uint16 if args.tif else np.uint8
    kcol = args.category_column
    ccol = args.color_table_category_name

    # Dangerous way to detect which columns to use as factor loadings
    with gzip.open(args.input, "rt") as rf:
        header = rf.readline().strip().split('\t')
    # Temporary - to be compatible with older input files
    recolumn = {'Hex_center_x':'x', 'Hex_center_y':'y', 'X':'x', 'Y':'y'}
    for i,x in enumerate(header):
        if x in recolumn:
            header[i] = recolumn[x]
    factor_header = []
    categorical = False
    if args.category_column != '':
        if args.category_column not in header:
            sys.exit(f"ERROR: {args.category_column} not found in header")
        categorical = True
        if not os.path.exists(args.color_table):
            sys.exit(f"ERROR: --color_table is required for categorical input")
        color_info = pd.read_csv(args.color_table, sep='\t', header=0, dtype={ccol:str})
        if args.input_rgb_uint8 or color_info[["R","G","B"]].max().max() > 2:
            for c in list("RGB"):
                color_info[c] = color_info[c] / 255
        color_info = color_info[~color_info[ccol].isna()]
        color_idx = {x:i for i,x in enumerate(color_info[ccol].values)}
        cmtx = np.array(color_info.loc[:, ["R","G","B"]])
        K = len(color_idx)
        factor_header = [str(k) for k in range(K)]
        logging.info(f"Use categorical input ({K} categories)")
    else:
        for x in header:
            y = re.match('^[A-Za-z]*_*(\d+)$', x)
            if y:
                factor_header.append([y.group(0), int(y.group(1)) ])
        factor_header.sort(key = lambda x : x[1] )
        factor_header = [x[0] for x in factor_header]
        K = len(factor_header)
        if os.path.exists(args.color_table):
            color_info = pd.read_csv(args.color_table, sep='\t', header=0)
            if args.input_rgb_uint8 or color_info[["R","G","B"]].max().max() > 2:
                for c in list("RGB"):
                    color_info[c] = color_info[c] / 255
            cmtx = np.array(color_info.loc[:, ["R","G","B"]])
        else:
            cmap_name = args.cmap_name
            if args.cmap_name not in plt.colormaps():
                cmap_name = "turbo"
            cmap = plt.get_cmap(cmap_name, K)
            cmtx = np.array([cmap(i) for i in range(K)] )
            indx = np.arange(K)
            shuffle(indx)
            cmtx = cmtx[indx, ]
            cmtx = cmtx[:, :3]
            cdict = {k:cmtx[k,:] for k in range(K)}
            # Plot color bar separately
            fig = plot_colortable(cdict, "Factor label", sort_colors=False, ncols=4)
            f = args.output + ".cbar"
            fig.savefig(f, format="png")
            logging.info(f"Set up color map for {K} factors")

    # Read data
    adt={x:float for x in ["x", "y"]+factor_header}
    adt[kcol] = str
    df = pd.DataFrame()
    for chunk in pd.read_csv(gzip.open(args.input, 'rt'), sep='\t', \
        chunksize=1000000, skiprows=1, names=header, dtype=adt):
        if args.scale > 0:
            chunk.x = chunk.x / args.scale
            chunk.y = chunk.y / args.scale
        chunk = chunk[(chunk.y > args.ymin) & (chunk.y < args.ymax)]
        chunk = chunk[(chunk.x > args.xmin) & (chunk.x < args.xmax)]
        chunk['x_indx'] = np.round(chunk.x.values / args.plot_um_per_pixel, 0).astype(int)
        chunk['y_indx'] = np.round(chunk.y.values / args.plot_um_per_pixel, 0).astype(int)
        if categorical: # Add dummy columns
            chunk[kcol] = chunk[kcol].map(color_idx).astype(int)
            for k in range(K):
                chunk[str(k)] = chunk[kcol].eq(k).astype(int)
        chunk = chunk.groupby(by = ['x_indx', 'y_indx']).agg({ x:np.mean for x in factor_header }).reset_index()
        df = pd.concat([df, chunk])

    df = df.groupby(by = ['x_indx', 'y_indx']).agg({ x:np.mean for x in factor_header }).reset_index()
    x_indx_min = args.xmin / args.plot_um_per_pixel
    y_indx_min = args.ymin / args.plot_um_per_pixel
    if args.plot_fit or np.isinf(args.xmin):
        df.x_indx = df.x_indx - int(np.max([x_indx_min, df.x_indx.min()]) )
    else:
        df.x_indx -= int(x_indx_min)
    if args.plot_fit or np.isinf(args.ymin):
        df.y_indx = df.y_indx - int(np.max([y_indx_min, df.y_indx.min()]) )
    else:
        df.y_indx -= int(y_indx_min)

    N0 = df.shape[0]
    df.index = range(N0)
    hsize, wsize = df[['x_indx','y_indx']].max(axis = 0) + 1
    hsize_um = hsize * args.plot_um_per_pixel
    wsize_um = wsize * args.plot_um_per_pixel
    logging.info(f"Read region {N0} pixels in region {hsize_um} x {wsize_um}")


    # Make images
    wst = df.y_indx.min()
    wed = df.y_indx.max()
    wstep = np.max([10, int(args.batch_size)])
    radius = args.fill_range/args.plot_um_per_pixel
    print(wsize, wstep)


    if radius <= 0:
        pts = df[['x_indx', 'y_indx']].values
        pts_indx = df.index.values
    else:
        pts = np.zeros((0, 2), dtype=int)
        pts_indx = []
        st = wst
        while st < wed:
            ed = min([st + wstep, wed])
            logging.info(f"Filling pixels {st} - {ed} / {wed}")
            if ((df.y_indx > st) & (df.y_indx < ed)).sum() < 10:
                st = ed
                continue
            block = df.index[(df.y_indx > st - 2*radius) & (df.y_indx < ed + 2*radius)]
            ref = sklearn.neighbors.BallTree(np.array(df.loc[block, ['x_indx', 'y_indx']]))
            mesh = np.meshgrid(np.arange(hsize), np.arange(st, ed))
            nodes = np.array(list(zip(*(dim.flat for dim in mesh))), dtype=int)
            dv, iv = ref.query(nodes, k = 1, dualtree=True)
            indx = dv[:, 0] < radius
            iv = block[iv[indx, 0] ]
            if sum(indx) == 0:
                st = ed
                continue
            pts = np.vstack((pts, nodes[indx, :]) )
            pts_indx += list(iv)
            st = ed
            if args.debug:
                break

    # Note: PIL default origin is upper-left
    pts[:,0] = np.clip(hsize - pts[:, 0], 0, hsize-1)
    pts[:,1] = np.clip(pts[:, 1], 0, wsize-1)
    if args.origin[0] > 0:
        pts[:,0] = hsize - 1 - pts[:, 0]
    if args.origin[1] > 0:
        pts[:,1] = wsize - 1 - pts[:, 1]

    logging.info(f"Start constructing RGB image")

    if not args.skip_mixture_plot:
        rgb_mtx = np.clip(np.around(np.array(df.loc[pts_indx,factor_header]) @\
                                    cmtx * 255),0,255).astype(dt)
        img = np.zeros( (hsize, wsize, 3), dtype=dt)
        for r in range(3):
            img[:, :, r] = coo_array((rgb_mtx[:, r], (pts[:,0], pts[:,1])),\
                shape=(hsize, wsize), dtype = dt).toarray()
        if args.tif:
            img = Image.fromarray(img, mode="I;16")
        else:
            img = Image.fromarray(img)

        outf = args.output
        outf += ".tif" if args.tif else ".png"
        img.save(outf)
        logging.info(f"Made fractional image\n{outf}")

    if args.plot_discretized:
        kvec = np.array(df.loc[pts_indx,factor_header]).argmax(axis = 1)
        cmtx = np.clip(np.around(cmtx * 255), 0, 255).astype(dt)
        img = np.zeros( (hsize, wsize, 3), dtype=dt)
        for r in range(3):
            img[:, :, r] = coo_array((cmtx[kvec, r], (pts[:,0], pts[:,1])),\
                shape=(hsize, wsize), dtype = dt).toarray()
        if args.tif:
            img = Image.fromarray(img, mode="I;16")
        else:
            img = Image.fromarray(img)

        outf = args.output + ".top"
        outf += ".tif" if args.tif else ".png"
        img.save(outf)
        logging.info(f"Made hard threshold image\n{outf}")

    if args.plot_individual_factor:
        if args.binary_cmap_name not in plt.colormaps():
            args.binary_cmap_name = "plasma"
        v = np.array(df.loc[pts_indx, factor_header].sum(axis = 0) )
        u = np.argsort(-v)
        for k in u:
            v = np.clip(df.loc[pts_indx, factor_header[k]].values,0,1)
            rgb_mtx = np.clip(mpl.colormaps[args.binary_cmap_name](v)[:,:3]*255,0,255).astype(dt)
            img = np.zeros( (hsize, wsize, 3), dtype=dt)
            for r in range(3):
                img[:, :, r] = coo_array((rgb_mtx[:, r], (pts[:,0], pts[:,1])),\
                    shape=(hsize, wsize), dtype = dt).toarray()
            if args.tif:
                img = Image.fromarray(img, mode="I;16")
            else:
                img = Image.fromarray(img)
            outf = args.output + ".F_"+str(k)
            outf += ".tif" if args.tif else ".png"
            img.save(outf)
            logging.info(f"Made factor specific image - {k}\n{outf}")

if __name__ == "__main__":
    plot_base(sys.argv[1:])



from scipy import sparse
import numpy as np

class Minibatch:

    def __init__(self, mtx, features = None, barcodes = None):
        assert sparse.issparse(mtx), "Invalid matrix - must be sparse"
        self.mtx = mtx.tocsr()
        self.n, self.M = mtx.shape
        self.gamma = None # n x K: P(k|n)
        self.alpha = None # Prior for gamma
        self.ll = 0       # log likelihood
        if features is None:
            self.feature = list(range(self.M))
        else:
            self.feature = features
        if barcodes is None:
            self.barcode = list(range(self.n))
        else:
            self.barcode = barcodes

    def init_from_matrix(self, gamma = None, alpha = None):
        if gamma is not None:
            assert gamma.shape[0] == self.n and len(gamma.shape) == 2, "Inconsistent dimensions of gamma"
            self.gamma = gamma
        if alpha is not None:
            self.alpha = alpha

class PairedMinibatch:

    def __init__(self, mtx_focal, mtx_buffer, buffer_weight = None, features = None, barcodes = None):
        assert sparse.issparse(mtx_focal), "Invalid matrix - must be sparse"
        assert sparse.issparse(mtx_buffer), "Invalid matrix - must be sparse"
        self.mtx = mtx_focal.tocsr()
        self.mtx_buffer = mtx_buffer.tocsr()
        self.n, self.M = self.mtx.shape
        self.gamma = None # n x K: P(k|n)
        self.gamma_buffer = None
        self.alpha = None # Prior for gamma
        self.ll = 0       # log likelihood
        if buffer_weight is None:
            self.buffer_weight = np.ones(self.n) * .5
        else:
            self.buffer_weight = np.array(buffer_weight).reshape(-1)
        if features is None:
            self.feature = list(range(self.M))
        else:
            self.feature = features
        if barcodes is None:
            self.barcode = list(range(self.n))
        else:
            self.barcode = barcodes

    def init_from_matrix(self, gamma = None, gamma_buffer = None, alpha = None):
        if gamma is not None:
            assert gamma.shape[0] == self.n and len(gamma.shape) == 2, "Inconsistent dimensions of gamma"
            self.gamma = gamma
        if gamma_buffer is not None:
            assert gamma_buffer.shape[0] == self.n and len(gamma_buffer.shape) == 2, "Inconsistent dimensions of gamma_buffer"
            self.gamma_buffer = gamma_buffer
        if alpha is not None:
            self.alpha = alpha



import sys, os, warnings, copy
import numpy as np
import pandas as pd
import sklearn.neighbors
import sklearn.mixture

from ficture.utils.hexagon_fn import pixel_to_hex, hex_to_pixel


def filter_by_density_mixture(df, key, radius, n_move, args):
    '''
    df: dataframe with columns X, Y, and key, (X, Y) are in um
    Return a dataframe with kept anchor point coordinates x, y
    '''
    pt = pd.DataFrame()
    m0v=[]
    m1v=[]
    hex_area = radius**2 * np.sqrt(3) * 3 / 2
    for i in range(n_move):
        for j in range(n_move):
            cnt = pd.DataFrame()
            cnt["hex_x"], cnt["hex_y"] = pixel_to_hex(np.asarray(df.loc[:, ['X','Y']]), radius, i/n_move, j/n_move)
            cnt[key] = copy.copy(df[key].values)
            cnt = cnt.groupby(by = ['hex_x','hex_y']).agg({key:"sum"}).reset_index()
            cnt = cnt[cnt[key] > hex_area * args.min_abs_mol_density_squm]
            if cnt.shape[0] < 10:
                continue
            if args.hard_threshold > 0:
                cnt['det'] = cnt[key] > hex_area * args.hard_threshold
            else:
                vorg = cnt[key].values
                if len(vorg) > args.max_npts_to_fit_model:
                    vorg = np.random.choice(vorg, int(args.max_npts_to_fit_model), replace=False)
                print(f"[{i}, {j}], {len(vorg)} units")
                # 1st try: fit a 2-component mixture model to log transformed density
                v = np.log10(vorg).reshape(-1, 1)
                gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v)
                lab_keep = np.argmax(gm.means_.squeeze())
                cnt['det'] = gm.predict(v) == lab_keep
                m0=(10**gm.means_.squeeze()[lab_keep])/hex_area
                m1=(10**gm.means_.squeeze()[1-lab_keep])/hex_area
                kept_min = cnt.loc[cnt.det.eq(True), key].min() / hex_area
                print(f"1st: log, 2 component. {m0:.3f} v.s. {m1:.3f}, kept min density {kept_min:.3f}")
                # If it does not seem right
                # 2nd try: fit a 3-component mixture model to log transformed density
                if m1 > m0 * 0.5 or m0 < args.min_abs_mol_density_squm_dense:
                    v = np.log10(vorg)
                    gm = sklearn.mixture.GaussianMixture(n_components=3, random_state=0).fit(v)
                    lab_rank = np.argsort(gm.means_.squeeze())
                    lab_keep = lab_rank[-1]
                    cnt['det'] = gm.predict(v) == lab_keep
                    m0=(10**gm.means_.squeeze()[lab_keep])/hex_area
                    m1=(10**gm.means_.squeeze()[lab_rank[1]])/hex_area
                    m2=(10**gm.means_.squeeze()[lab_rank[0]])/hex_area
                    kept_min = cnt.loc[cnt.det.eq(True), key].min() / hex_area
                    print(f"2nd: log, 3 component. {m0:.3f} v.s. {m1:.3f} & {m2:.3f}, kept min density {kept_min:.3f}")
                # # 3rd try: fit a 2-component mixture model to density of original scale
                # if m1 > m0 * 0.5 or m0 < args.min_abs_mol_density_squm_dense:
                #     v = vorg.reshape(-1, 1)
                #     gm = sklearn.mixture.GaussianMixture(n_components=2, random_state=0).fit(v)
                #     lab_keep = np.argmax(gm.means_.squeeze())
                #     cnt['det'] = gm.predict(v) == lab_keep
                #     m0=gm.means_.squeeze()[lab_keep]/hex_area
                #     m1=gm.means_.squeeze()[1-lab_keep]/hex_area
                #     kept_min = cnt.loc[cnt.det.eq(True), key].min() / hex_area
                #     print(f"2nd: 2 component. {m0:.3f} v.s. {m1:.3f}, kept min density {kept_min:.3f}")
                if m0 < args.min_abs_mol_density_squm_dense:
                    cnt['det'] = cnt[key] > hex_area * args.min_abs_mol_density_squm_dense
                elif m1 > m0 * .7:
                    cnt['det'] = 1
                m0v.append(m0)
                m1v.append(m1)
            if cnt.det.eq(True).sum() < 2:
                continue
            m0 = cnt.loc[cnt.det.eq(True), key].median()/hex_area
            m1 = cnt.loc[cnt.det.eq(False), key].median()/hex_area
            m0v.append(m0)
            m1v.append(m1)
            anchor_x, anchor_y = hex_to_pixel(cnt.loc[cnt.det.eq(True), 'hex_x'].values, cnt.loc[cnt.det.eq(True), 'hex_y'].values, radius,i/n_move,j/n_move)
            pt = pd.concat([pt,\
                    pd.DataFrame({'x':anchor_x, 'y':anchor_y})])
    return pt, np.mean(m0v), np.mean(m1v)



import sys, os, gzip, copy, gc, logging
import argparse
import numpy as np
import pandas as pd
from scipy.sparse import *
import subprocess as sp
import geojson
import shapely.prepared, shapely.geometry
from shapely.geometry import Point

# Add parent directory
#sys.path.append(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
#from hexagon_fn import *
from ficture.utils.hexagon_fn import pixel_to_hex, hex_to_pixel

def make_sge_by_hexagon(_args):

    parser = argparse.ArgumentParser(prog = "make_sge_by_hexagon")
    parser.add_argument('--input', type=str, help='')
    parser.add_argument('--feature', type=str, help='')
    parser.add_argument('--output_path', type=str, help='')
    parser.add_argument('--boundary', type=str, default = '', help='')

    parser.add_argument('--major_axis', type=str, default="Y", help='X or Y')
    parser.add_argument('--mu_scale', type=float, default=26.67, help='Coordinate to um translate')
    parser.add_argument('--key', default = 'gn', type=str, help='gt: genetotal, gn: gene, spl: velo-spliced, unspl: velo-unspliced')
    parser.add_argument('--precision', type=int, default=1, help='Number of digits to store spatial location (in um), 0 for integer.')
    parser.add_argument('--hex_width', type=int, default=24, help='')
    parser.add_argument('--hex_radius', type=int, default=-1, help='')
    parser.add_argument('--overlap', type=float, default=-1, help='')
    parser.add_argument('--n_move', type=int, default=1, help='')
    parser.add_argument('--min_ct_density', type=float, default=0.1, help='Minimum density of output hexagons, in nUMI/um^2')
    parser.add_argument('--min_ct_per_unit', type=float, default=-1, help='Minimum umi count of output hexagons')
    parser.add_argument('--transfer_gene_prefix', action="store_true", help='')
    args = parser.parse_args(_args)

    if len(_args) == 0:
        parser.print_help()
        return

    if not os.path.exists(args.input):
        sys.exit(f"ERROR: cannot find input file \n {args.input}")
    logging.basicConfig(level= getattr(logging, "INFO", None))

    mu_scale = 1./args.mu_scale
    radius=args.hex_radius
    b_size = 512

    diam=args.hex_width
    radius=args.hex_radius
    if radius < 0:
        radius = diam / np.sqrt(3)
    else:
        diam = int(radius*np.sqrt(3))
    if args.overlap >= 0 and args.overlap < 1:
        n_move = int(1 / (1. - args.overlap) )
    else:
        n_move = args.n_move
        if n_move < 0:
            n_move = 1

    hex_area = diam*radius*3/2

    min_ct_per_unit = args.min_ct_per_unit
    if args.min_ct_per_unit < 0:
        min_ct_per_unit = args.min_ct_density * hex_area

    ### Output
    if not os.path.exists(args.output_path):
        arg="mkdir -p " + args.output_path
        os.system(arg)

    with gzip.open(args.input, 'rt') as rf:
        input_header=rf.readline().strip().split('\t')

    feature=pd.read_csv(args.feature,sep='\t',header=0,usecols=['gene', 'gene_id', args.key])
    feature.sort_values(by=args.key, ascending=False, inplace=True)
    feature.drop_duplicates(subset='gene', inplace=True)
    if args.transfer_gene_prefix:
        prefix = feature.gene.map(lambda x: x.split('_')[0] + '_' if '_' in x else '').values
        feature.gene_id = prefix + feature.gene_id.values
    feature['dummy'] = "Gene Expression"
    f = args.output_path + "/features.tsv.gz"
    feature[['gene_id','gene','dummy']].to_csv(f, sep='\t', index=False, header=False)

    feature_kept = list(feature.gene.values)
    ft_dict = {x:i for i,x in enumerate( feature_kept ) }
    M = len(feature_kept)

    brc_f = args.output_path + "/barcodes.tsv"
    mtx_f = args.output_path + "/matrix.mtx"
    # If exists, delete
    if os.path.exists(brc_f):
        _ = os.system("rm " + brc_f)
    if os.path.exists(mtx_f):
        _ = os.system("rm " + mtx_f)

    use_boundary = False
    if os.path.isfile(args.boundary):
        try:
            mpoly = shapely.geometry.shape(geojson.load(open(args.boundary, 'rb')))
        except:
            gj = geojson.load(open(args.boundary, 'rb'))
            mpoly = shapely.geometry.shape(gj['features'][0]['geometry'] )
        mpoly = shapely.prepared.prep(mpoly)
        use_boundary = True
        logging.info(f"Load boundary from {args.boundary}")

    n_unit = 0
    T = 0

    adt = {x:float for x in ['X','Y', args.key]}
    adt.update({x:str for x in ['gene', 'gene_id']})
    df_buff = pd.DataFrame()
    for chunk in pd.read_csv(gzip.open(args.input, 'rt'),sep='\t',chunksize=1000000, header=0, usecols=['X','Y','gene','gene_id',args.key], dtype=adt):
        if chunk.shape[0] == 0:
            break
        ed = chunk[args.major_axis].iloc[-1]
        left = copy.copy(chunk[chunk[args.major_axis].gt(ed - 5 * args.mu_scale)])
        df_buff = pd.concat([df_buff, chunk])
        df_buff['j'] = df_buff.X.astype(str) + '_' + df_buff.Y.astype(str)
        df = df_buff[df_buff.gene.isin(ft_dict)]
        brc_full = df.groupby(by = ['j','X','Y']).agg({args.key: sum}).reset_index()
        brc_full.index = range(brc_full.shape[0])
        brc_full['crd'] = None
        pts = np.asarray(brc_full[['X','Y']]) * mu_scale
        logging.info(f"Read data with {brc_full.shape[0]} pixels and {feature.shape[0]} genes.")

        # Make DGE
        barcode_kept = list(brc_full.j.values)
        bc_dict = {x:i for i,x in enumerate( barcode_kept ) }
        indx_row = [ bc_dict[x] for x in df['j']]
        indx_col = [ ft_dict[x] for x in df['gene']]
        N = len(barcode_kept)

        dge_mtx = coo_matrix((df[args.key].values, (indx_row, indx_col)), shape=(N, M)).tocsr()
        total_molecule=df[args.key].sum()
        logging.info(f"Made DGE {dge_mtx.shape}")

        offs_x = 0
        offs_y = 0
        while offs_x < n_move:
            while offs_y < n_move:
                x,y = pixel_to_hex(pts, radius, offs_x/n_move, offs_y/n_move)
                brc_full['crd'] = list(zip(x,y))
                hex_info = brc_full.groupby(by = 'crd').agg({args.key: sum})
                hex_info.rename(columns = {args.key:'tot'}, inplace=True)
                hex_info = hex_info.loc[hex_info.tot > args.min_ct_per_unit, :]
                if use_boundary:
                    hex_coord = np.array([[x,y] for x,y in hex_info.index.values])
                    x, y = hex_to_pixel(hex_coord[:, 0], hex_coord[:, 1], radius, offs_x/n_move, offs_y/n_move)
                    kept = [mpoly.contains(Point(*p)) for p in zip(x, y)]
                    logging.info(f"Keep {sum(kept)}/{len(x)} units within boundary.")
                    hex_info = hex_info.loc[kept, :]

                hex_id = set(hex_info.index.values)
                if len(hex_id) < 1:
                    offs_y += 1
                    continue
                hex_list = list(hex_id)
                hex_dict = {x:i for i,x in enumerate(hex_list)}

                # Hexagon and pixel correspondence
                sub = pd.DataFrame({'crd':brc_full.crd.values,'cCol':range(N)})
                sub = sub[sub.crd.isin(hex_id)]
                sub['cRow'] = sub.crd.map(hex_dict).astype(int)

                # Hexagon level info (for barcodes.tsv.gz)
                brc = brc_full[brc_full.crd.isin(hex_id)].groupby(by = 'crd').agg({'X':np.mean, 'Y':np.mean, args.key:sum}).reset_index()
                brc['cRow'] = brc.crd.map(hex_dict).astype(int)
                brc['X'] = [f"{x:.{args.precision}f}" for x in brc.X.values]
                brc['Y'] = [f"{x:.{args.precision}f}" for x in brc.Y.values]
                brc.sort_values(by = 'cRow', inplace=True)
                with open(brc_f, 'a') as wf:
                    _ = wf.write('\n'.join(\
                    (brc.cRow+n_unit+1).astype(str).values + '_' + \
                    str(offs_x) + '.' + str(offs_y) + '_' + \
                    brc.X.values + '_' + brc.Y.values + '_' + \
                    brc[args.key].astype(str).values) + '\n')
                n_hex = len(hex_dict)
                n_minib = n_hex // b_size + 1
                mid_ct = np.median(hex_info.tot.values)
                logging.info(f"{offs_x}, {offs_y}: {n_minib}, {n_hex} ({sub.cRow.max()}, {sub.shape[0]}), median count per unit {mid_ct}")

                # Output sparse matrix
                grd_minib = list(range(0, n_hex, b_size))
                grd_minib.append(n_hex)
                st_minib = 0
                n_minib = len(grd_minib) - 1
                while st_minib < n_minib:
                    indx_minib = (sub.cRow >= grd_minib[st_minib]) & (sub.cRow < grd_minib[st_minib+1])
                    npixel_minib = sum(indx_minib)
                    offset = sub.loc[indx_minib, 'cRow'].min()
                    nhex_minib = sub.loc[indx_minib, 'cRow'].max() - offset + 1
                    mtx = coo_matrix((np.ones(npixel_minib, dtype=bool), (sub.loc[indx_minib, 'cRow'].values-offset, sub.loc[indx_minib, 'cCol'].values)), shape=(nhex_minib, N) ).tocsr() @ dge_mtx
                    mtx.eliminate_zeros()
                    r, c = mtx.nonzero()
                    r = np.array(r,dtype=int) + offset + n_unit + 1
                    c = np.array(c,dtype=int) + 1
                    T += len(r)
                    mtx = pd.DataFrame({'i':c, 'j':r, 'v':mtx.data})
                    mtx['i'] = mtx.i.astype(int)
                    mtx['j'] = mtx.j.astype(int)
                    mtx['v'] = mtx.v.astype(int)
                    mtx.to_csv(mtx_f, mode='a', sep=' ', index=False, header=False)
                    #print(mtx.shape[0])
                    st_minib += 1
                    if ( st_minib % 50 == 0):
                        logging.info(f"{st_minib}/{n_minib}. Wrote {nhex_minib} units.")
                n_unit += brc.shape[0]
                logging.info(f"Sliding offset {offs_x}, {offs_y}. Wrote {n_unit} units so far.")
                offs_y += 1
            offs_y = 0
            offs_x += 1
        df_buff = copy.copy(left)

    _ = os.system("gzip -f " + brc_f)

    mtx_header = args.output_path + "/matrix.header"
    with open(mtx_header, 'w') as wf:
        line = "%%MatrixMarket matrix coordinate integer general\n%\n"
        line += " ".join([str(x) for x in [M, n_unit, T]]) + "\n"
        wf.write(line)

    arg = " ".join(["cat",mtx_header,mtx_f,"| gzip -c > ", mtx_f+".gz"])
    if os.system(arg) == 0:
        _ = os.system("rm " + mtx_f)
        _ = os.system("rm " + mtx_header)

if __name__ == "__main__":
    make_sge_by_hexagon(sys.argv[1:])


import sys, io, os, gzip, glob, copy, re, time, pickle, warnings, importlib
import numpy as np
import pandas as pd
import base64

from sklearn.preprocessing import normalize
# from sklearn.decomposition import TruncatedSVD, PCA
import scipy.cluster
import scipy.stats
import scipy.spatial
import scipy.linalg

import ete3
from ete3 import Tree, TreeStyle, NodeStyle
os.environ['QT_QPA_PLATFORM']='offscreen'

def logrank(x):
    v = scipy.stats.rankdata(x)
    return - np.log( 1-(v-.5)/len(v) )

def cor_logrank(orgmtx):
    K, M = orgmtx.shape
    rankmtx = np.zeros((K, M), dtype=int)
    for k in range(K):
        rankmtx[k, :] = scipy.stats.rankdata(orgmtx[k, :])
    corlogrank = np.corrcoef(-np.log( 1-(rankmtx-.5)/M ) )
    return corlogrank

def NJ_logrank(orgmtx):
    K, M = orgmtx.shape
    rankmtx = np.zeros((K, M), dtype=int)
    for k in range(K):
        rankmtx[k, :] = scipy.stats.rankdata(orgmtx[k, :])
    corlogrank = np.corrcoef(-np.log( 1-(rankmtx-.5)/M ) )
    node = {} # node id -> [leaves, profile]
    active_node = []
    for k in range(K):
        node[k] = [[k], orgmtx[k, :]]
        active_node.append(k)
    pairwise_dict = {} # (id1, id2) -> score
    for k in range(K-1):
        for l in range(k+1, K):
            pairwise_dict[(k, l)] = corlogrank[k, l]
    Z = []
    for s in range(K-1):
        candi = [] # [(id1, id2), score]
        for i,v1 in enumerate(active_node[:-1]):
            for j in range(i+1, len(active_node)):
                v2 = active_node[j]
                if (v1, v2) not in pairwise_dict:
                    vec1 = logrank(node[v1][1])
                    vec2 = logrank(node[v2][1])
                    score = np.corrcoef(vec1, vec2)[0,1]
                    candi.append([(v1, v2), score])
                    pairwise_dict[ (v1, v2) ] = score
                else:
                    score = pairwise_dict[(v1, v2)]
                    candi.append([(v1, v2), score])
        if len(candi) == 0:
            break
        candi.sort(key = lambda x : x[1], reverse=True)
        v1, v2 = candi[0][0]
        lvs = node[v1][0]+node[v2][0]
        node[K+s] = [lvs, orgmtx[lvs, :].mean(axis = 0)]
        active_node.remove(v1)
        active_node.remove(v2)
        active_node.append(K+s)
        Z.append([ v1, v2, 1-candi[0][1], len(lvs) ])
    return Z

def visual_hc(model_prob, weight, top_gene, node_color=None, factor_name=None, circle=False, vertical=False, output_f=None, cprob_cut=.99):

    K = model_prob.shape[0]
    assert len(weight) == K, "model_prob.shape[0] != len(weight)"
    assert len(top_gene) == K, "len(top_gene) != K"
    if factor_name is None:
        factor_name = [str(x) for x in range(K)]

    model_prob = normalize(np.array(model_prob), norm='l1', axis=1)
    weight = np.array(weight)
    weight /= weight.sum()
    weight_anno = ["%.2e" % x if x < 0.1 else "%.3f" % x for x in weight]
    v = np.argsort(weight)[::-1]
    w = np.cumsum(weight[v] )
    if sum(w > cprob_cut) == 0:
        k = K - 1
    else:
        k = np.arange(K)[w > cprob_cut][0]
    kept_factor = factor_name[v[:(k+1)]]
    kept_idx = v[:(k+1)].astype(str)

    # # Hierarchical clustering
    # cd_dist = scipy.spatial.distance.pdist(model_prob, metric='cosine')
    # Z_hc = scipy.cluster.hierarchy.linkage(cd_dist, method="complete")
    # Z_hc = NJ_logrank(model_prob)

    corlogrank = cor_logrank(model_prob)
    corlogrank = np.nan_to_num(corlogrank, copy=False)
    cd_dist = 1 - .5 * (corlogrank + corlogrank.T)
    np.fill_diagonal(cd_dist, 0)
    cd_dist = scipy.spatial.distance.squareform(cd_dist)
    Z_hc = scipy.cluster.hierarchy.linkage(cd_dist, method="complete")

    # Construct tree object from the clustering
    R, T = scipy.cluster.hierarchy.to_tree(Z_hc, rd=True)
    tr = Tree()
    tr.dist=0
    tr.name='root'
    node_dict = {R.id:tr}
    stack = [R]
    while stack:
        node = stack.pop()
        c_dist = node.dist / 2
        for c in [node.left, node.right]:
            if c:
                ch = Tree()
                ch.dist = c_dist
                ch.name = str(c.id)
                _=node_dict[node.id].add_child(ch)
                node_dict[c.id] = ch
                stack.append(c)

    node_list = [x for x in tr.traverse() if x.is_leaf() and x.name in
    kept_idx]
    subtr = tr.copy()
    subtr.prune( [x.name for x in node_list] )
    for x in subtr.traverse():
        if x.is_leaf():
            x.name = factor_name[int(x.name)]

    if output_f is None:
        return subtr

    ### Visualize tree
    title=f"Hierarchical clustering of {len(node_list)} factors"

    # Node style
    istyle = NodeStyle()
    istyle["size"] = 0
    for n in subtr.traverse():
        n.set_style(istyle)
    if node_color is not None:
        for n in subtr.traverse():
            if n.is_leaf():
                nstyle = NodeStyle()
                nstyle["fgcolor"] = node_color[n.name]
                nstyle['size'] = 10
                n.set_style(nstyle)
    node_anno = {factor_name[k]: " " + factor_name[k] + " ("+weight_anno[k] + "): " + v  for k,v in enumerate(top_gene) }
    def layout(node):
        if node.is_leaf():
            ete3.faces.add_face_to_node(ete3.TextFace(node_anno[node.name]), node, column=0)

    # Tree style
    ts = TreeStyle()
    ts.layout_fn=layout
    ts.show_leaf_name = False
    ts.show_branch_length = False
    ts.show_scale = False
    if circle:
        ts.mode = "c"
        ts.arc_start = 0
        ts.arc_span = 360
    else:
        if vertical:
            ts.rotation = 90
        ts.branch_vertical_margin = 20

    ts.title.add_face(ete3.TextFace(title, fsize=25),column=0)
    subtr.render(output_f, w=2560, units='mm', tree_style=ts)

    return subtr

def image_to_base64(image_path):
    with open(image_path, "rb") as image_file:
        return base64.b64encode(image_file.read()).decode("utf-8")
